/*****************************************************************************\
 
   Name . . . . . : DTSPLITM.C
   
   Author . . . . : Paul Markham (3:711/934.1@fidonet)
   
   Date written . : April 1994
   
 ------------------------------------------------------------------------------
   This code is released to the public domain. You may use and/or modified it
   in any way you choose.
  
   There is no warranty of any kind associated with this code. If you use it,
   you accept all responsibility for any loss or damages that may occur.
 ------------------------------------------------------------------------------
 
  Split a fidonet or seadog date into its component parts. See dtsplitm.h 
  for details.
        
\*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include "dtsplitm.h"

/*---------------------------------------------------------------------------*\
   Defines
\*---------------------------------------------------------------------------*/

#define TRUE                1
#define FALSE               0

#define TOKEN_NUM           10
#define TOKEN_SIZE          10

#define SEA_MIN_TOKEN       6
#define FIDO_MIN_TOKEN      6

#define FIDO_TOKEN_DAY      0
#define FIDO_TOKEN_MONTH    1
#define FIDO_TOKEN_YEAR     2
#define FIDO_TOKEN_HOUR     3
#define FIDO_TOKEN_MIN      4
#define FIDO_TOKEN_SEC      5

#define SEA_TOKEN_DAY       1
#define SEA_TOKEN_MONTH     2
#define SEA_TOKEN_YEAR      3
#define SEA_TOKEN_HOUR      4
#define SEA_TOKEN_MIN       5

#define CT_JUNK             0
#define CT_NUM              1
#define CT_ALPHA            2

#define MIN_HOUR            0
#define MAX_HOUR            23
#define MIN_MIN             0
#define MAX_MIN             59
#define MIN_SEC             0
#define MAX_SEC             59
#define MIN_YEAR            0
#define MAX_YEAR            99
#define MIN_DAY             1
#define MAX_DAY             31

#define MIN_LIKENESS        3

/*---------------------------------------------------------------------------*\
   Globals
\*---------------------------------------------------------------------------*/

static char *monthTab[]
    = {"jan", "feb", "mar", "apr" ,"may", "jun",
       "jul", "aug", "sep", "oct", "nov", "dec"};
#define MONTH_NUM  (sizeof(monthTab) / sizeof(monthTab[0]))

/*---------------------------------------------------------------------------*\
   Function prototypes
\*---------------------------------------------------------------------------*/

static int  charType(char c);
static int  fidoDate(char token[TOKEN_NUM][TOKEN_SIZE],
                     int *year, int *month, int *day,
                     int *hour, int *min, int *sec);
static int  guessMonth(char *s, int *month);
static int  isNumStr(char *s);
static int  matchExactMonth(char *s, int *month);
static int  seaDate(char token[TOKEN_NUM][TOKEN_SIZE],
                    int *year, int*month, int *day,
                    int *hour, int *min, int *sec);
static int  tokeniseDate(char *dateTime, char token[TOKEN_NUM][TOKEN_SIZE]);
static int  validateYear(char *s, int *year);
static int  validateMonth(char *s, int *month);
static int  validateDay(char *s, int *day);
static void validateHour(char *s, int *hour);
static void validateMin(char *s, int *min);
static void validateSec(char *s, int *sec);

/*---------------------------------------------------------------------------*\
   Date spliting routine
\*---------------------------------------------------------------------------*/

int dtsplitm(const char *date, int *year, int *month, int *day,
           int *hour, int *min, int *sec)
    {
    char    dateTime[50];
    char    token[TOKEN_NUM][TOKEN_SIZE];
    int     numToken;
    int     rc;
    unsigned char *p;

    strcpy(dateTime, date);
    p = (unsigned char *)dateTime;
    while (*p != '\0')
    {
        *p = (unsigned char)tolower(*p);
        p++;
    }
    memset(token, '\0', sizeof(token));

    numToken = tokeniseDate(dateTime, token);
    if (isalpha(token[0][0]))
        {
        if (numToken < SEA_MIN_TOKEN)
            rc = DATE_ERROR;
        else
            rc = seaDate(token, year, month, day, hour, min, sec);
        }
    else
        {
        if (numToken < FIDO_MIN_TOKEN)
            rc = DATE_ERROR;
        else
            rc = fidoDate(token, year, month, day, hour, min, sec);
        }

    return rc;
    }

/*---------------------------------------------------------------------------*\
   Convert date into a series of tokens. A token is delimited by a character
   that is not a number or an alpha, or a character that is of a different
   type - eg. a number is delimited by an alpha character.  
\*---------------------------------------------------------------------------*/

static int tokeniseDate(char *dateTime, char token[TOKEN_NUM][TOKEN_SIZE])
    {
    char    currToken[TOKEN_SIZE];
    int     prevCharType;
    int     currCharType;
    int     dtIdx;
    int     ctIdx;
    int     tokIdx;
    int     done;

    prevCharType = charType(dateTime[0]);
    dtIdx = 0;
    ctIdx = 0;
    tokIdx = 0;
    done = FALSE;

    do
        {
        currCharType = charType(dateTime[dtIdx]);
        if (dateTime[dtIdx] == '\0')
            done = TRUE;
        if (currCharType != prevCharType)
            {
            currToken[ctIdx] = '\0';
            ctIdx = 0;
            if (prevCharType != CT_JUNK)
                {
                strcpy(token[tokIdx], currToken);
                tokIdx++;
                }
            }
        currToken[ctIdx] = dateTime[dtIdx];
        ctIdx++;
        dtIdx++;
        prevCharType = currCharType;
        }
    while(!done);

    return tokIdx;
    }

/*---------------------------------------------------------------------------*\
   Treat the tokenised date as a fido format date and try to figure out
   what the date and time is.
\*---------------------------------------------------------------------------*/

static int  fidoDate(char token[TOKEN_NUM][TOKEN_SIZE],
                     int *year, int*month, int *day,
                     int *hour, int *min, int *sec)
    {
    int rc;

    validateHour(token[FIDO_TOKEN_HOUR], hour);
    validateMin(token[FIDO_TOKEN_MIN], min);
    validateSec(token[FIDO_TOKEN_SEC], sec);

    rc = validateDay(token[FIDO_TOKEN_DAY], day);
    if (rc == DATE_OK)
        {
        rc = validateMonth(token[FIDO_TOKEN_MONTH], month);
        if (rc == DATE_OK)
            rc = validateYear(token[FIDO_TOKEN_YEAR], year);
        }

    return rc;
    }

/*---------------------------------------------------------------------------*\
   Treat the tokenised date as a SEA format date and try to figure out what 
   the date and time is.
\*---------------------------------------------------------------------------*/

static int  seaDate(char token[TOKEN_NUM][TOKEN_SIZE],
                    int *year, int*month, int *day,
                    int *hour, int *min, int *sec)
    {
    int rc;

    validateHour(token[SEA_TOKEN_HOUR], hour);
    validateMin(token[SEA_TOKEN_MIN], min);
    *sec = 0;
    rc = validateDay(token[SEA_TOKEN_DAY], day);
    if (rc == DATE_OK)
        {
        rc = validateMonth(token[SEA_TOKEN_MONTH], month);
        if (rc == DATE_OK)
            rc = validateYear(token[SEA_TOKEN_YEAR], year);
        }

    return rc;
    }

/*---------------------------------------------------------------------------*\
   Validate the hour. If the hour isn't valid change to something valid.
\*---------------------------------------------------------------------------*/

static void validateHour(char *s, int *hour)
    {
    *hour = atoi(s);
    if (*hour < MIN_HOUR)
        *hour = MIN_HOUR;
    else
        if (*hour > MAX_HOUR)
            *hour = MAX_HOUR;
    }

/*---------------------------------------------------------------------------*\
   Validate the minute. If the minutes isn't valid change it to something 
   valid.
\*---------------------------------------------------------------------------*/

static void validateMin(char *s, int *min)
    {
    *min = atoi(s);
    if (*min < MIN_MIN)
        *min = MIN_MIN;
    else
        if (*min > MAX_MIN)
            *min = MAX_MIN;
    }

/*---------------------------------------------------------------------------*\
   Validate the second. If the second isn't valid change it something valid. 
\*---------------------------------------------------------------------------*/

static void validateSec(char *s, int *sec)
    {
    *sec = atoi(s);
    if (*sec < MIN_SEC)
        *sec = MIN_SEC;
    else
        if (*sec > MAX_SEC)
            *sec = MAX_SEC;
    }

/*---------------------------------------------------------------------------*\
   Validate the day. A non numeric day will return an error. If the day is 
   numeric but isn't valid change it to something valid.
\*---------------------------------------------------------------------------*/

static int validateDay(char *s, int *day)
    {
    int rc;

    rc = DATE_OK;

    if (isNumStr(s))
        {
        *day = atoi(s);
        if (*day < MIN_DAY)
            *day = MIN_DAY;
        else
            if (*day > MAX_DAY)
                *day = MAX_DAY;
        }
    else
        rc = DATE_ERROR;

    return rc;
    }                  
    
/*---------------------------------------------------------------------------*\
   Validate the month. If an exact match can't be found, try to figure out
   what it should be anyway.
\*---------------------------------------------------------------------------*/

static int validateMonth(char *s, int *month)
    {
    int rc;

    rc = matchExactMonth(s, month);
    if (rc != DATE_OK)
        rc = guessMonth(s, month);

    return rc;
    }

/*---------------------------------------------------------------------------*\
   Find the month string that exactly matches the one passed. 
\*---------------------------------------------------------------------------*/

static int matchExactMonth(char *s, int *month)
    {
    int rc;
    int i;

    rc = DATE_ERROR;

    for (i = 0; i < MONTH_NUM; i++)
        if (strcmp(s, monthTab[i]) == 0)
            {
            *month = i + 1;
            rc = DATE_OK;
            break;
            }

    return rc;
    }

/*---------------------------------------------------------------------------*\
   Take the best guess we can as to what the month is. This is done by
   comparing the passed month to the list of valid months. A count is
   incremented for each leading character that matches as well as each
   character the matches throughout the whole string. The month with the
   highest count is the winner. If the resulting count is over the minimum
   specified in MIN_LIKENESS then consider it a match, otherwise return an
   error.
\*---------------------------------------------------------------------------*/

static int guessMonth(char *s, int *month)
    {
    int rc;
    int i;
    int j;
    int monthLen;
    int likeness;
    int bestLikeness;

    rc = DATE_ERROR;
    monthLen = strlen(monthTab[0]);
    if (strlen(s) < monthLen)
    {
        monthLen = strlen(s);
    }
    bestLikeness = 0;

    for (i = 0; i < MONTH_NUM; i++)
        {
        likeness = 0;
        for (j = 0; j < monthLen; j++)
            if (s[j] == monthTab[i][j])
                likeness++;
            else
                break;

        for (j = 0; j < monthLen; j++)
            if (s[j] == monthTab[i][j])
                likeness++;

        if (likeness >= MIN_LIKENESS)
            if (likeness > bestLikeness)
                {
                rc = DATE_OK;
                bestLikeness = likeness;
                *month = i + 1;
                }
        }
    return rc;
    }

/*---------------------------------------------------------------------------*\
   Validate the year. A non numeric year will return an error. If the year is 
   numeric, figure out whether it is supposed to be this cenury or the next.
\*---------------------------------------------------------------------------*/

static int validateYear(char *s, int *year)
    {
    int         rc;
    time_t      t;
    struct tm   *tm;

    rc = DATE_OK;
    t = time(NULL);
    tm = localtime(&t);

    if (isNumStr(s))
        {
        *year = atoi(s);
        if (*year < MIN_YEAR || *year > MAX_YEAR)
            rc = DATE_ERROR;
        else
            {
            *year += 1900;
            if (abs(1900 + tm->tm_year - *year) > 50)
                *year += 100;
            }
        }
    else
        rc = DATE_ERROR;

    return rc;
    }

/*---------------------------------------------------------------------------*\
   Return type of character. Anything that isn't a numeric or alpha will be
   considered to be junk.
\*---------------------------------------------------------------------------*/

static int charType(char c)
    {
    int     rc;

    if (isdigit(c))
        rc = CT_NUM;
    else
        if (isalpha(c))
            rc = CT_ALPHA;
        else
            rc = CT_JUNK;

    return rc;
    }

/*---------------------------------------------------------------------------*\
   Return TRUE if the string contains only numeric charcters. 
\*---------------------------------------------------------------------------*/

static int isNumStr(char *s)
    {
    int rc;
    int i;

    rc = TRUE;

    for (i = 0; i < strlen(s); i++)
        if (!isdigit(s[i]))
            {
            rc = FALSE;
            break;
            }

    return rc;
    }
