/* mailproc - mailprocessor */
/* Work started by Paul Edwards, 1993-12-30 */
/* Released to the Public Domain */

/* This set of routines is designed to do the mailprocessing proper step
   of Tobruk.  */


#include <stdlib.h>
#include <time.h>

#ifdef USE_MSGAPI
#include <msgapi.h>
#endif

#include "mailproc.h"
#include "error.h"
#include "estdlib.h"
#include "dtsplit.h"
#include "dtsplitm.h"
#include "dtfido.h"
#include "exists.h"
#include "reform.h"
#include "findkblk.h"
#include "unused.h"
#include "patmat.h"

static void mailprocNodeSetup(MAILPROC *mailproc);
static void mailprocHandlePacket(MAILPROC *mailproc);
static void mailprocHandleMessage(MAILPROC *mailproc);
static void mailprocParseIntl(MAILPROC *mailproc, unsigned char *intl);
static void mailprocProcessEchomail(MAILPROC *mailproc);
static void mailprocProcessEchomail2(MAILPROC *mailproc);
static void mailprocScourSeenby(MAILPROC *mailproc);
static void mailprocScourPath(MAILPROC *mailproc);
static void mailprocScourOrigin(MAILPROC *mailproc);
static void mailprocHandleMessage2(MAILPROC *mailproc);
static void mailprocHandleMessage3(MAILPROC *mailproc);
static void mailprocHandleMessage4(MAILPROC *mailproc);
static void mailprocProcessNetmail(MAILPROC *mailproc);
static void mailprocGateCheck(MAILPROC *mailproc);
static void mailprocRemap(MAILPROC *mailproc, unsigned char *to);
static void mailprocRoute(MAILPROC *mailproc, fidoAddress *fa);
static void mailprocIterateDests(MAILPROC *mailproc,
                                 void (*destHandler)(MAILPROC *mailproc));
                                 
static void mailprocSendEchomail(MAILPROC *mailproc);
static void mailprocWriteEchoText(MAILPROC *mailproc);
static void mailprocWriteEchoText2(MAILPROC *mailproc);
static void mailprocWriteSeenby(MAILPROC *mailproc);
static void mailprocWritePath(MAILPROC *mailproc);

static void mailprocSendNetmail(MAILPROC *mailproc,
                                void (*handler)(MAILPROC *mailproc));
static void mailprocSendNetmail2(MAILPROC *mailproc);
static void mailprocWriteNetmailText(MAILPROC *mailproc);
static void mailprocWriteIntl(MAILPROC *mailproc);
static void mailprocWriteVia(MAILPROC *mailproc);

static void mailprocCurrentArea(MAILPROC *mailproc);
static void mailprocGetDests(MAILPROC *mailproc);
static int fidoAddressCompare(fidoAddress *fa1, fidoAddress *fa2);
static int miniAddressCompare(fidoAddress *fa1, fidoAddress *fa2);
static void mailprocNewPacket(MAILPROC *mailproc);
static void mailprocPacketHeader(MAILPROC *mailproc);
static void mailprocClosePackets(MAILPROC *mailproc);
static void mailprocMessageBase(MAILPROC *mailproc);
static void mailprocMsgOrigin(MAILPROC *mailproc);

#ifdef USE_MBASE
static void mailprocMsgOut(MAILPROC *mailproc);
static void mailprocMsgWrite(MAILPROC *mailproc, GMD *msg);
static void mailprocFetchMessage3(MAILPROC *mailproc);
static void mailprocScourEndMsg(MAILPROC *mailproc);
static void mailprocCalcSeenby(MAILPROC *mailproc);
static void mailprocCalcPath(MAILPROC *mailproc);
static void mailprocNetmailMsgSetup(MAILPROC *mailproc);
static void mailprocWriteMsgHeader(MAILPROC *mailproc);
static void mailprocWriteAreaLine(MAILPROC *mailproc);
static void mailprocWriteKludges(MAILPROC *mailproc);
static void mailprocWriteTextBody(MAILPROC *mailproc);
#endif

#ifdef USE_MSGAPI
static void mailprocMsgWrite(MAILPROC *mailproc, MSGH *msgh, XMSG *msg);
static void mailprocFetchMessage(MAILPROC *mailproc);
static void mailprocFetchMessage2(MAILPROC *mailproc);
static void mailprocFetchMessage3(MAILPROC *mailproc);
static void mailprocNetmailMsgSetup(MAILPROC *mailproc);
static void mailprocGetMsgHeader(MAILPROC *mailproc);
static void mailprocUpdateMsgHeader(MAILPROC *mailproc);
static void mailprocScourEndMsg(MAILPROC *mailproc);
static void mailprocFindEndMsg(MAILPROC *mailproc);
static void mailprocCalcSeenby(MAILPROC *mailproc);
static void mailprocCalcPath(MAILPROC *mailproc);
static void mailprocMsgOut(MAILPROC *mailproc);
static void mailprocWriteMsgHeader(MAILPROC *mailproc);
static void mailprocWriteAreaLine(MAILPROC *mailproc);
static void mailprocWriteKludges(MAILPROC *mailproc);
static void mailprocWriteTextBody(MAILPROC *mailproc);
#endif

void mailprocDefaults(MAILPROC *mailproc)
{
    mailproc->seenbyMax = MAILPROC_SEENBY_MAX;
    mailproc->pathMax = MAILPROC_PATH_MAX;
    mailproc->maxControl = sizeof mailproc->tempControl - 1;
    pktDefaults(&mailproc->pkt);
#ifdef USE_MBASE    
    mbaseDefaults(&mailproc->mbase);
#endif    
    mailproc->mainbufSize = 10000;
    mailproc->inbufsize = 0;
    mailproc->inbuf = NULL;
    return;
}

void mailprocInit(MAILPROC *mailproc)
{
#ifdef USE_MSGAPI
    int x;
#endif    
    
    mailproc->mainAddr = mailproc->baseAddr;
    seenbyInit(&mailproc->seenby, mailproc->seenbyMax);
    seenbyInit(&mailproc->path, mailproc->pathMax);
    pktSetSize(&mailproc->pkt, mailproc->mainbufSize);
    mailproc->mnbufsz = mailproc->mainbufSize - 1;
    if ((mailproc->mnbufsz % 2) != 0)
    {
        mailproc->mnbufsz--;
        mailproc->mnbufhsz = mailproc->mnbufsz / 2;
    }
    mailproc->mainbuf = emalloc(mailproc->mainbufSize);
    if (ALLOK)
    {
        pktSetBuf(&mailproc->pkt, mailproc->mainbuf);
        pktInit(&mailproc->pkt);
    }
#ifdef USE_MBASE
    if (ALLOK)
    {
        mbaseProvideAreas(&mailproc->mbase, 
                         mailproc->areas,
                         mailproc->numAreas,
                         &mailproc->netarea);
        mbaseInit(&mailproc->mbase);
        if (ALLOK)
        {
            mbaseGetAreaList(&mailproc->mbase, &mailproc->aset);
            if (!ALLOK)
            {
                mbaseTerm(&mailproc->mbase);
            }
        }
    }
#endif    
#ifdef USE_MSGAPI    
    if (ALLOK)
    {
        mailproc->minf.req_version = 0;
        mailproc->minf.def_zone = (word)mailproc->baseAddr.zone;
        if (MsgOpenApi(&mailproc->minf) != 0)
        {
            errorSet(MSGAPI_OPEN, msgapierr);
        }
        else
        {
            for (x=0; x < mailproc->numAreas; x++)
            {
                mailproc->areas[x].mh = NULL;
            }
            mailproc->netarea.mh = NULL;
        }
    }
#endif    
    if (ALLOK)
    {
        if (mailproc->inbufsize != 0)
        {
            mailproc->inbuf = emalloc(mailproc->inbufsize);
            if (ALLOK)
            {
                pktSetFileBuf(&mailproc->pkt, 
                              mailproc->inbuf, 
                              mailproc->inbufsize);
            }
        }
        if (ALLOK)
        {
            mailprocNodeSetup(mailproc);
            mailprocReset(mailproc);
        }
    }
    return;
}

void mailprocReset(MAILPROC *mailproc)
{
    mailproc->debug = 0;
    mailproc->okCount = 0;
    mailproc->route = NULL;
    mailproc->baseAddr = mailproc->mainAddr;
    strcpy(mailproc->address, "");
    strcpy(mailproc->packet, "");
    strcpy(mailproc->netarea.messageBase, "");
    mailproc->sequence = 0;
    mailproc->destCount = 0;
    mailproc->totalmsgs = 0;
    mailproc->totalmsgsOut = 0;
    strcpy(mailproc->lastArea, "");
    mailproc->lastAreaLen = 1;
    time(&mailproc->startTime);
    return;
}

void mailprocTerm(MAILPROC *mailproc)
{
#ifdef USE_MSGAPI
    int x;
#endif    
    
    mailprocClosePackets(mailproc);
#ifdef USE_MBASE
    mbaseTerm(&mailproc->mbase);
#endif        
#ifdef USE_MSGAPI
    for (x = 0; x < mailproc->numAreas; x++)
    {
        if (mailproc->areas[x].mh != NULL)
        {
            MsgCloseArea(mailproc->areas[x].mh);
        }
    }    
    if (mailproc->netarea.mh != NULL)
    {
        MsgCloseArea(mailproc->netarea.mh);
    }
    MsgCloseApi();
#endif
    return;
}

void mailprocPacket(MAILPROC *mailproc, char *packetName)
{
    char *newName;
    
    mailproc->msgnum = 0;
    strcpy(mailproc->packet, packetName);
    pktSetFile(&mailproc->pkt, mailproc->packet);
    if (ALLOK)
    {
        mailprocHandlePacket(mailproc);
        if (!ALLOK)
        {
            errorFlush();
            errorSet(MAILPROC_BADMSG, mailproc->msgnum);
        }
        mailproc->totalmsgs += mailproc->msgnum;
        pktFinFile(&mailproc->pkt);
        if (ALLOK)
        {
            eremove(mailproc->packet);
        }
        else if (*mailproc->badName != '\0')
        {
            errorFlush();
            newName = reform(mailproc->packet, 
                             mailproc->goodName,
                             mailproc->badName);
            if (newName == NULL)
            {
                errorSet(MAILPROC_RENAME, 
                         mailproc->packet,
                         mailproc->goodName,
                         mailproc->badName);
            }
            else
            {
                erename(mailproc->packet, newName);
            }
        }
    }
    return;                         
}

static void mailprocNodeSetup(MAILPROC *mailproc)
{
    unused(mailproc);
    return;
}

static void mailprocHandlePacket(MAILPROC *mailproc)
{
    raw_fido_packet_hdr raw_p;
    fido_packet_hdr fp_hdr;    

    memcpy(&raw_p, mailproc->pkt.startPtr, sizeof raw_p);
    if (ALLOK)           
    {
        cvt_from_raw_packet_hdr(&raw_p, &fp_hdr);
        if (mailproc->debug)
        {
            printf("packet going from %d:%d/%d.%d to %d:%d/%d.%d\n",
                   fp_hdr.origZone,
                   fp_hdr.origNet, 
                   fp_hdr.origNode,
                   fp_hdr.origPoint,
                   fp_hdr.destZone,
                   fp_hdr.destNet, 
                   fp_hdr.destNode,
                   fp_hdr.destPoint);
            printf("packet created on %.4d-%.2d-%.2d %.2d:%.2d:%.2d\n",
                   fp_hdr.pdate.year,
                   fp_hdr.pdate.month + 1,
                   fp_hdr.pdate.day,
                   fp_hdr.pdate.hour,
                   fp_hdr.pdate.min,
                   fp_hdr.pdate.sec);
        }
        strcpy(mailproc->pktOrigin.domain, mailproc->baseAddr.domain);
        mailproc->pktOrigin.zone = fp_hdr.origZone;
        mailproc->pktOrigin.net = fp_hdr.origNet;
        mailproc->pktOrigin.node = fp_hdr.origNode;
        mailproc->pktOrigin.point = fp_hdr.origPoint;
        strcpy(mailproc->pktDest.domain, mailproc->baseAddr.domain);
        mailproc->pktDest.zone = fp_hdr.destZone;
        mailproc->pktDest.net = fp_hdr.destNet;
        mailproc->pktDest.node = fp_hdr.destNode;
        mailproc->pktDest.point = fp_hdr.destPoint;
        /*
        if (!fidoAddressCompare(&mailproc->pktDest, &mailproc->baseAddr))
        {
            errorSet(MAILPROC_PKT_NOT_MINE, mailproc->packet);
        }
        */
        if (ALLOK)
        {
            pktFirstMsg(&mailproc->pkt);
            while (ALLOK && !mailproc->pkt.endOfPkt)
            {
                mailprocHandleMessage(mailproc);
                if (ALLOK)
                {
                    pktNextMsg(&mailproc->pkt);
                }
            }
        }
    }
    return;
}

static void mailprocHandleMessage(MAILPROC *mailproc)
{
    memcpy(&mailproc->raw_m, 
           mailproc->pkt.startPtr + 2, 
           sizeof mailproc->raw_m);
    if (ALLOK)
    {         
        mailproc->msgnum++;         
        if (mailproc->debug)
        {
             printf("message %d:\n", mailproc->msgnum);
        }
        mailprocHandleMessage2(mailproc);
    }
    return;
}

static void mailprocHandleMessage2(MAILPROC *mailproc)
{
    cvt_from_raw_message_hdr(&mailproc->raw_m, &mailproc->fm_hdr);
    strcpy(mailproc->msgOrigin.domain, mailproc->baseAddr.domain);
    mailproc->msgOrigin.zone = mailproc->baseAddr.zone;
    mailproc->msgOrigin.net = mailproc->fm_hdr.origNet;
    mailproc->msgOrigin.node = mailproc->fm_hdr.origNode;
    mailproc->msgOrigin.point = 0;
    strcpy(mailproc->msgDest.domain, mailproc->baseAddr.domain);
    mailproc->msgDest.zone = mailproc->baseAddr.zone;
    mailproc->msgDest.net = mailproc->fm_hdr.destNet;
    mailproc->msgDest.node = mailproc->fm_hdr.destNode;
    mailproc->msgDest.point = 0;
    if (mailproc->debug)
    {
        printf("date: %s\n", mailproc->pkt.date);
        printf("from %d/%d to %d/%d\n",
               mailproc->fm_hdr.origNet, 
               mailproc->fm_hdr.origNode,
               mailproc->fm_hdr.destNet,
               mailproc->fm_hdr.destNode); 
    }
    mailprocHandleMessage3(mailproc);
    return;
}

static void mailprocHandleMessage3(MAILPROC *mailproc)
{
    if (ALLOK)
    {
        if (mailproc->debug) 
        {
            printf("from %s to %s subject %s\n",
                   mailproc->pkt.from,
                   mailproc->pkt.to,
                   mailproc->pkt.subject);
        }
        mailprocHandleMessage4(mailproc);
    }
    return;
}
            
static void mailprocHandleMessage4(MAILPROC *mailproc)
{
    mailproc->echoFlag = 0;
    if (memcmp(mailproc->pkt.text,
               "\x41\x52\x45\x41\x3a",
               5) == 0)
    {
        memcpy(mailproc->conference, 
               mailproc->pkt.text + 5,
               sizeof mailproc->conference);               
        mailproc->conference[sizeof mailproc->conference - 1] = '\0';        
        mailproc->tptr = 
            (unsigned char *)strchr((char *)mailproc->conference, 0x0d);
        if (mailproc->tptr != NULL)
        {
            *mailproc->tptr = '\0';
        }
        if (*mailproc->conference == ' ')
        {
            memmove(mailproc->conference,
                    mailproc->conference+1,
                    strlen((char *)(mailproc->conference+1))+1);
        }        
        if (mailproc->debug)
        {
            printf("conference %s\n", mailproc->conference);
        }
        mailproc->echoFlag = 1;
        mailprocProcessEchomail(mailproc);
    }
    else
    {
        mailprocProcessNetmail(mailproc);
    }
    return;
}    

static void mailprocProcessEchomail(MAILPROC *mailproc)
{
    if ((mailproc->msgOrigin.net != mailproc->pktOrigin.net)
        || (mailproc->msgOrigin.node != mailproc->msgOrigin.node))
    {
        errorSet(MAILPROC_ECHOMAIL_ORIGIN);
    }
    if (ALLOK)
    {
        if ((mailproc->msgDest.net != mailproc->pktDest.net)
            || (mailproc->msgDest.node != mailproc->pktDest.node))
        {
            errorSet(MAILPROC_ECHOMAIL_DEST);
        }
    }
    if (ALLOK)
    {
        mailprocProcessEchomail2(mailproc);
    }
    return;
}

static void mailprocProcessNetmail(MAILPROC *mailproc)
{
    mailproc->recipAddr = mailproc->msgDest;
    mailproc->authorAddr = mailproc->msgOrigin;
    mailproc->tptr = 
        (unsigned char *)strstr((char *)mailproc->pkt.text, "\x01" "FMPT ");
    if (mailproc->tptr != NULL)
    {
        mailproc->authorAddr.point = atoi((char *)mailproc->tptr + 6);
    }
    mailproc->tptr = 
        (unsigned char *)strstr((char *)mailproc->pkt.text, "\x01" "TOPT ");
    if (mailproc->tptr != NULL)
    {
        mailproc->recipAddr.point = atoi((char *)mailproc->tptr + 6);
    }
    mailproc->tptr = 
        (unsigned char *)strstr((char *)mailproc->pkt.text, "\x01" "INTL ");
    if (mailproc->tptr != NULL)
    {
        mailprocParseIntl(mailproc, mailproc->tptr + 6);
    }
    mailprocGateCheck(mailproc);
    mailproc->currAreaDef = &mailproc->netarea;
    
    /* remapped is used to determine whether the message had it's
       recipient address changed, and thus needs an new INTL
       and/or TOPT written */
       
    mailproc->remapped = 0;
    
    /* routed is used to determine whether the message was meant for
       this address, or whether it was sent on to someone else */
       
    mailproc->routed = 0;
    
    /* only check for remaps if the message is specifically destined
       for this node */
       
    if (fidoAddressCompare(&mailproc->recipAddr, &mailproc->baseAddr))
    {
        mailprocRemap(mailproc, mailproc->pkt.to);
    }
    
    /* if the 2d address is not mine, then route it elsewhere,
         only looking at the 2d address.
       otherwise, if the 5d recipient address is not my address,
         then route using the 5d address */
    
    if (!miniAddressCompare(&mailproc->msgDest, &mailproc->baseAddr))
    {
        mailprocRoute(mailproc, &mailproc->msgDest);
    }
    if (!fidoAddressCompare(&mailproc->recipAddr, &mailproc->baseAddr))
    {
        mailprocRoute(mailproc, &mailproc->recipAddr);
    }
    if (mailproc->routed)
    {
        mailprocSendNetmail(mailproc, mailprocSendNetmail2);
        mailproc->totalmsgsOut++;
    }
    mailprocMessageBase(mailproc);    
    return;
}

static void mailprocGateCheck(MAILPROC *mailproc)
{
    /* any messages from points should be gated */
    if (mailproc->baseAddr.point == 0)
    {
        if ((mailproc->authorAddr.point != 0)
            && (mailproc->authorAddr.node == mailproc->baseAddr.node)
            && (mailproc->authorAddr.net == mailproc->baseAddr.net)
            && (mailproc->authorAddr.zone == mailproc->baseAddr.zone)
            && (strcmp(mailproc->authorAddr.domain, 
                       mailproc->baseAddr.domain) == 0))
        {
            if (mailproc->recipAddr.zone != mailproc->baseAddr.zone)
            {
                mailproc->msgDest.net = mailproc->baseAddr.zone;
                mailproc->msgDest.node = mailproc->recipAddr.zone;
            }
        }
    }
    
    /* any messages to me should be gated */
    if (fidoAddressCompare(&mailproc->msgDest,
                           &mailproc->baseAddr))
    {
        if (!fidoAddressCompare(&mailproc->msgDest, &mailproc->recipAddr))
        {
            mailproc->msgDest = mailproc->recipAddr;
        }
    }
    return;
}

static void mailprocRemap(MAILPROC *mailproc, unsigned char *to)
{
    int x;
    
    for (x = 0; x < mailproc->remapCount; x++)
    {
        if (patmat((char *)to, mailproc->remaps[x].name))
        {
            mailproc->recipAddr = mailproc->remaps[x].addr;
            mailproc->msgDest = mailproc->remaps[x].addr;
            mailproc->remapped = 1;
            break;
        }  
    }
    return;
}

static void mailprocRoute(MAILPROC *mailproc, fidoAddress *fa)
{
    int y;

    mailproc->routed = 0;
    
    /* if we have a routing file available, use it */
    
    if (mailproc->route != NULL)
    {
        if (routeMap(mailproc->route, fa, &mailproc->routeAddr))
        {
            mailproc->routed = 1;
        }
    }
    
    /* otherwise, see if it is a known destination */
    
    if (!mailproc->routed)
    {
        for (y = 0; y < mailproc->destCount; y++)
        {
            if (fidoAddressCompare(fa, &mailproc->dests[y].addr))
            {
                mailproc->routeAddr = mailproc->dests[y].addr;
                mailproc->routed = 1;
                break;
            }
        }
    }
    
    /* if I am a point, then route stuff to my boss */
    if (!mailproc->routed)
    {
        if (mailproc->baseAddr.point != 0)
        {
            strcpy(mailproc->routeAddr.domain, mailproc->baseAddr.domain);
            mailproc->routeAddr.zone = mailproc->baseAddr.zone;
            mailproc->routeAddr.net = mailproc->baseAddr.net;
            mailproc->routeAddr.node = mailproc->baseAddr.node;
            mailproc->routeAddr.point = 0;
            mailproc->routed = 1;
        }
    }
    /* now worldroute anything else, unless it is for a point
       off my system, then SWALLOW the message */
    if (!mailproc->routed)
    {
        if ((mailproc->baseAddr.point == 0)
            && (fa->point != 0)
            && (mailproc->baseAddr.zone == fa->zone)
            && (mailproc->baseAddr.net == fa->net)
            && (mailproc->baseAddr.node == fa->node)
            && (strcmp(fa->domain, mailproc->baseAddr.domain) == 0))
        {
            mailproc->routed = 0;
        }
        else
        {
            mailproc->routeAddr = mailproc->worldRoute;
            mailproc->routed = 1;
        }
    }
    return;
}

static void mailprocParseIntl(MAILPROC *mailproc, unsigned char *intl)
{
    FADDR fadd_to;
    FADDR fadd_fr;
    fidoAddress to;
    fidoAddress from;
    
    while (*intl == 0x20) intl++;
    faddrInit(&fadd_to);
    faddrCreate(&fadd_to, (char *)intl);
    intl = (unsigned char *)strchr((char *)intl, 0x20);
    if (intl != NULL)
    {
        intl++;
        faddrInit(&fadd_fr);
        faddrCreate(&fadd_fr, (char *)intl);
        faddrGetaddr(&fadd_to, &to);
        faddrGetaddr(&fadd_fr, &from);
        mailproc->authorAddr.zone = from.zone;
        mailproc->authorAddr.net = from.net;
        mailproc->authorAddr.node = from.node;
        mailproc->recipAddr.zone = to.zone;
        mailproc->recipAddr.net = to.net;
        mailproc->recipAddr.node = to.node;
    }
    return;
}

static void mailprocProcessEchomail2(MAILPROC *mailproc)
{
    if (mailproc->pkt.endPtr == NULL)
    {
        pktEnd(&mailproc->pkt);
    }
    if (ALLOK)
    {
        mailprocCurrentArea(mailproc);
        if (ALLOK)
        {
            mailproc->baseAddr = mailproc->currAreaDef->primary;
            mailproc->recipAddr = mailproc->baseAddr;
            mailprocScourSeenby(mailproc);
            if (ALLOK)
            {
                mailprocScourPath(mailproc);
                if (ALLOK)
                {
                    mailprocScourOrigin(mailproc);
                    if (mailproc->pkt.startPtr == NULL)
                    {
                        pktBegin(&mailproc->pkt);
                    }
                    if (ALLOK)
                    {
                        mailprocGetDests(mailproc);
                        if (ALLOK)
                        {
                            seenbySort(&mailproc->seenby);
                            mailprocIterateDests(mailproc, 
                                                 mailprocSendEchomail);
                        }
                    }
                }
            }
            mailproc->baseAddr = mailproc->mainAddr;    
        }
        else if (errorCompare(MAILPROC_UNKNOWN_AREA))
        {
            errorFlush();
        }
    }
    return;
}

static void mailprocScourSeenby(MAILPROC *mailproc)
{
    unsigned char *top;
    unsigned char *p;
    int ret;
    
    seenbyReset(&mailproc->seenby);
    if (mailproc->pkt.startPtr == NULL)
    {
        top = mailproc->pkt.startValT + 100; /* 100 bytes safety */
    }
    else
    {
        top = mailproc->pkt.text; /* was startPtr; */
    }
    ret = findkblk(top,
                   mailproc->pkt.endPtr,
                   &mailproc->startSeenby,
                   &mailproc->endSeenby,
                   &mailproc->numSeenby,
                   (unsigned char *)"SEEN-BY:",
                   8);
    if (ret == 1)
    {
        errorSet(MAILPROC_NO_SEENBY);
    }
    else if ((ret == 2) && (mailproc->pkt.startPtr == NULL))
    {
        errorSet(MAILPROC_TOO_MANY_SEENBY);
    }
    else
    {
        p = mailproc->startSeenby;
        while ((p != NULL) && (p < mailproc->endSeenby))
        {
            seenbyAddline(&mailproc->seenby, p + 8);
            p = (unsigned char *)strchr((char *)p, 0x0d);
            if (p != NULL)
            {
                p++;
                if (*p == 0x0a)
                {
                    p++;
                }
            }
        }
    }
    return;
}
    
static void mailprocScourPath(MAILPROC *mailproc)
{
    unsigned char *top;
    unsigned char *p;
    int ret;
    
    seenbyReset(&mailproc->path);
    if (mailproc->pkt.startPtr == NULL)
    {
        top = mailproc->pkt.startValT + 100; /* 100 bytes safety */
    }
    else
    {
        top = mailproc->pkt.startPtr;
    }
    ret = findkblk(top,
                   mailproc->pkt.endPtr,
                   &mailproc->startPath,
                   &mailproc->endPath,
                   &mailproc->numPath,
                   (unsigned char *)"\x01" "PATH:",
                   6);
    if (ret == 2)
    {
        errorSet(MAILPROC_TOO_MANY_PATH);
    }
    else if (ret == 0)
    {
        p = mailproc->startPath;
        while ((p != NULL) && (p < mailproc->endPath))
        {
            seenbyAddline(&mailproc->path, p + 6);
            p = (unsigned char *)strchr((char *)p, 0x0d);
            if (p != NULL)
            {
                p++;
                if (*p == 0x0a)
                {
                    p++;
                }
            }
        }
    }
    if (ret != 2)
    {
        seenbyAddNode(&mailproc->path, &mailproc->baseAddr);
    }
    return;
}

/* Look for origin line just above seenby, otherwise look
   anywhere for it */
        
static void mailprocScourOrigin(MAILPROC *mailproc)
{
    unsigned char *top;
    unsigned char *p;
    unsigned char *oldp;
    int foundOrigin;
    
    mailproc->origin[0] = '\0';
    top = mailproc->pkt.startValT;
    oldp = mailproc->startSeenby;
    foundOrigin = 0;
    while ((oldp > top) && (*oldp != 0x0d))
    {
        oldp--;
    }
    p = oldp;
    oldp--;
    while ((oldp > top) && (*oldp != 0x0d))
    {
        oldp--;
    }
    if (oldp > top) 
    {
        oldp++;
        if (*oldp == 0x0a)
        {
            oldp++;
        }
        if (memcmp(oldp, " * Origin:", 10) == 0)
        {
            foundOrigin = 1;
        }
    }
    if (!foundOrigin)
    {
        p = top;
        do
        {
            oldp = p;
            p = (unsigned char *)strstr((char *)p + 1, " * Origin:");
        } while (p != NULL);
        if (oldp != top)
        {
            p = oldp - 1;
            if (*p == 0x0a)
            {
                p--;
            }
            if (*p == 0x0d)
            {
                p = (unsigned char *)strchr((char *)oldp, 0x0d);
                if (p != NULL)
                {
                    foundOrigin = 1;
                }
            }
        }
    }
    if (foundOrigin)
    {
        if ((p - oldp) >= sizeof mailproc->origin)
        {
            memcpy(mailproc->origin,
                   p - sizeof mailproc->origin + 1,
                   sizeof mailproc->origin - 1);
            mailproc->origin[sizeof mailproc->origin - 1] = '\0';
        }
        else
        {
            memcpy(mailproc->origin, oldp, (size_t)(p - oldp));
            mailproc->origin[(size_t)(p - oldp)] = '\0';
        }
    }
    mailprocMsgOrigin(mailproc);
    return;
}
    
static void mailprocIterateDests(MAILPROC *mailproc,
                                 void (*destHandler)(MAILPROC *mailproc))
{
    int x;
    int y;
    char *mode;

    for (x = 0; x < mailproc->currAreaDef->numNodes; x++)
    {
        for (y = 0; y < mailproc->destCount; y++)
        {
            if (fidoAddressCompare(&mailproc->currAreaDef->outList[x].fa, 
                                   &mailproc->dests[y].addr))
            {
                if (mailproc->currAreaDef->outList[x].process)
                {
                    mailproc->confAreaNum = x;
                    mailproc->confDestNum = y;
                    mailproc->currAddress =
                        &mailproc->dests[y].addr;
                    if (!mailproc->dests[y].packetOpened)
                    {
                        mailprocNewPacket(mailproc);
                        strcpy(mailproc->dests[y].packetName, 
                               mailproc->newPacket);
                        if (exists(mailproc->newPacket))
                        {
                            mode = "r+b";
                        }
                        else
                        {
                            mode = "wb";
                        }       
                        efileOpen(&mailproc->dests[y].packetFile,
                                  mailproc->newPacket,
                                  mode);
                        if (ALLOK)
                        {
                            mailproc->dests[y].packetOpened = 1;
                            mailproc->currDest = &mailproc->dests[y].packetFile;
                            if (strcmp(mode, "wb") == 0)
                            {
                                mailprocPacketHeader(mailproc);
                            }
                            else
                            {
                                efileSeek(&mailproc->dests[y].packetFile,
                                          -2L,
                                          SEEK_END);
                            }              
                        }
                    }
                    if (mailproc->dests[y].packetOpened)
                    {
                        mailproc->currDest = &mailproc->dests[y].packetFile;
                        destHandler(mailproc);
                    }
                }
                break;
            }
        }
    }
    return;
} 

static void mailprocMessageBase(MAILPROC *mailproc)
{
#if (!defined(USE_MSGAPI) && !defined(USE_MBASE))
    unused(mailproc);
    return;
#endif
#ifdef USE_MBASE
    GMD msg;
    GAH gah;
    GAD gad;
    
    mbaseAsetQueryByName(&mailproc->mbase, 
                         &mailproc->aset, 
                         &gah,
                         mailproc->currAreaDef->areaName);
    mbaseGetAreaData(&mailproc->mbase, &gah, &gad);
    if (gad.passThrough)
    {
        return;
    }
    if (mailproc->pkt.startPtr == NULL)
    {
        pktBegin(&mailproc->pkt);
    }
    if (ALLOK)
    {
        memset(&msg,'\0',sizeof msg);
        msg.attr = 0;
        if (!mailproc->echoFlag && mailproc->routed)
        {
            msg.attr |= ATTR_SENT;
        }
        msg.orig_4d.zone = mailproc->authorAddr.zone;
        msg.orig_4d.net = mailproc->authorAddr.net;
        msg.orig_4d.node = mailproc->authorAddr.node;
        msg.orig_4d.point = mailproc->authorAddr.point;
        if (!mailproc->echoFlag)
        {
            msg.attr |= ATTR_PRIVATE;
            if ((mailproc->recipAddr.zone != mailproc->baseAddr.zone)
                || (mailproc->recipAddr.net != mailproc->baseAddr.net)
                || (mailproc->recipAddr.node != mailproc->baseAddr.node)
                || (mailproc->recipAddr.point != mailproc->baseAddr.point))
            {
                msg.attr |= ATTR_TRANSIT;
            }
            msg.to_2d.zone = mailproc->recipAddr.zone;
            msg.to_2d.net = mailproc->recipAddr.net;
            msg.to_2d.node = mailproc->recipAddr.node;
            msg.to_2d.point = mailproc->recipAddr.point;
            msg.to_4d = msg.to_2d;
            msg.from_4d = msg.orig_4d;
            msg.from_2d = msg.from_4d;
        }
        else
        {
            msg.to_2d.zone = mailproc->baseAddr.zone;
            msg.to_2d.net = mailproc->baseAddr.net;
            msg.to_2d.node = mailproc->baseAddr.node;
            msg.to_2d.point = mailproc->baseAddr.point;
            msg.to_4d = msg.to_2d;
        }
        strncpy((char *)msg.from, 
                (char *)mailproc->pkt.from,
                sizeof msg.from);
        msg.from[sizeof msg.from - 1] = '\0';
        strncpy((char *)msg.to, 
                (char *)mailproc->pkt.to,
                sizeof msg.to);
        msg.to[sizeof msg.to - 1] = '\0';
        strncpy((char *)msg.subj,
                (char *)mailproc->pkt.subject,
                sizeof msg.subj);
        msg.subj[sizeof msg.subj - 1] = '\0';
        strncpy((char *)msg.create_time,
                (char *)mailproc->pkt.date,
                sizeof msg.create_time);
        msg.create_time[sizeof msg.create_time - 1] = '\0';
          
        msg.gah = &gah;

        if (ALLOK)
        {
            mailprocMsgWrite(mailproc, &msg);
            if (!ALLOK)
            {
                errorSet(MAILPROC_MSG_WRITE, mailproc->msgnum);
            }
        }
    }
    return;
#endif  
#ifdef USE_MSGAPI
    XMSG msg;
    MSGH *msgh;
    char area[50];
    time_t tt;
    struct tm *tms;
    word type;
    AREADEFN *current;
    
    current = mailproc->currAreaDef;
    if (current->passThrough)
    {
        return;
    }
    if (current->mbType == MAILPROC_MB_SQUISH)
    {
        type = MSGTYPE_SQUISH;
    }
    else
    {
        type = MSGTYPE_SDM;
        if (mailproc->echoFlag)
        {
            type |= MSGTYPE_ECHO;
        }
    }
    if (mailproc->pkt.startPtr == NULL)
    {
        pktBegin(&mailproc->pkt);
    }
    if (current->mh == NULL)
    {
        strcpy(area, mailproc->currAreaDef->messageBase);
  
        if ((current->mh =
             MsgOpenArea((byte *)area, 
                         MSGAREA_CRIFNEC, 
                         type)) == NULL)
        {
            errorSet(MSG_AREA_OPEN, area, msgapierr);
        }
    }
    if (ALLOK)
    {
        if ((msgh = MsgOpenMsg(current->mh, 
                               MOPEN_CREATE, 
                               0)) == NULL)
        {
            errorSet(MSG_MSG_CREATE, msgapierr);
        }
        else
        {
            memset(&msg,'\0',sizeof msg);
            msg.attr = 0;
            if (!mailproc->echoFlag && mailproc->routed)
            {
                msg.attr |= MSGSENT;
            }
            if (!mailproc->echoFlag)
            {
                msg.attr |= MSGPRIVATE;
                if ((mailproc->recipAddr.zone != mailproc->baseAddr.zone)
                    || (mailproc->recipAddr.net != mailproc->baseAddr.net)
                    || (mailproc->recipAddr.node != mailproc->baseAddr.node)
                    || (mailproc->recipAddr.point != mailproc->baseAddr.point))
                {
                    msg.attr |= MSGFWD;
                }
                msg.dest.zone = (word)mailproc->recipAddr.zone;
                msg.dest.net = (word)mailproc->recipAddr.net;
                msg.dest.node = (word)mailproc->recipAddr.node;
                msg.dest.point = (word)mailproc->recipAddr.point;
            }
            else
            {
                msg.dest.zone = (word)mailproc->baseAddr.zone;
                msg.dest.net = (word)mailproc->baseAddr.net;
                msg.dest.node = (word)mailproc->baseAddr.node;
                msg.dest.point = (word)mailproc->baseAddr.point;
            }
            strncpy((char *)msg.from, 
                    (char *)mailproc->pkt.from,
                    sizeof msg.from);
            msg.from[sizeof msg.from - 1] = '\0';
            strncpy((char *)msg.to, 
                    (char *)mailproc->pkt.to,
                    sizeof msg.to);
            msg.to[sizeof msg.to - 1] = '\0';
            strncpy((char *)msg.subj,
                    (char *)mailproc->pkt.subject,
                    sizeof msg.subj);
            msg.subj[sizeof msg.subj - 1] = '\0';
            dt_split(mailproc->pkt.date,
                     &mailproc->mdate.year,
                     &mailproc->mdate.month,
                     &mailproc->mdate.day,
                     &mailproc->mdate.hour,
                     &mailproc->mdate.min,
                     &mailproc->mdate.sec);
            if (!ALLOK && errorCompare(DTSPLIT_BAD_DATE))
            {
                errorFlush();
                if (dtsplitm((char *)mailproc->pkt.date,
                             &mailproc->mdate.year,
                             &mailproc->mdate.month,
                             &mailproc->mdate.day,
                             &mailproc->mdate.hour,
                             &mailproc->mdate.min,
                             &mailproc->mdate.sec) != 0)
                {
                    errorSet(MAILPROC_FUCKED_DATE, mailproc->pkt.date);
                    errorFlush();
                    mailproc->mdate.year = 1994;
                    mailproc->mdate.month = 1;
                    mailproc->mdate.day = 1;
                    mailproc->mdate.hour = 1;
                    mailproc->mdate.min = 1;
                    mailproc->mdate.sec = 1;
                }
            }
            msg.date_written.date.yr = mailproc->mdate.year - 1980;
            msg.date_written.date.mo = mailproc->mdate.month;
            msg.date_written.date.da = mailproc->mdate.day;
            msg.date_written.time.hh = mailproc->mdate.hour;
            msg.date_written.time.mm = mailproc->mdate.min;
            msg.date_written.time.ss = mailproc->mdate.sec;
            strncpy((char *)msg.__ftsc_date, 
                    (char *)mailproc->pkt.date,
                    sizeof msg.__ftsc_date);
            msg.__ftsc_date[sizeof msg.__ftsc_date - 1] = '\0';
            
            time(&tt);
            tms = localtime(&tt);
            msg.date_arrived.date.yr = tms->tm_year - 80;
            msg.date_arrived.date.mo = tms->tm_mon + 1;
            msg.date_arrived.date.da = tms->tm_mday;
            msg.date_arrived.time.hh = tms->tm_hour;
            msg.date_arrived.time.mm = tms->tm_min;
            msg.date_arrived.time.ss = tms->tm_sec; 
            msg.orig.zone = (word)mailproc->authorAddr.zone;
            msg.orig.net = (word)mailproc->authorAddr.net;
            msg.orig.node = (word)mailproc->authorAddr.node;
            msg.orig.point = (word)mailproc->authorAddr.point;

            if (ALLOK)
            {
                mailprocMsgWrite(mailproc, msgh, &msg);
                if (!ALLOK)
                {
                    errorSet(MAILPROC_MSG_WRITE, mailproc->msgnum);
                }
            }
  
            MsgCloseMsg(msgh);
        }
    }
    return;
#endif  
}

static void mailprocPacketHeader(MAILPROC *mailproc)
{
    fido_packet_hdr fp_hdr;
    raw_fido_packet_hdr raw_p;
    time_t tt;
    struct tm *tms;

    memset(&fp_hdr, '\0', sizeof fp_hdr);
    fp_hdr.origNode = mailproc->baseAddr.node;
    fp_hdr.destNode = mailproc->currAddress->node;
    time(&tt);
    tms = localtime(&tt);
    fp_hdr.pdate.year = tms->tm_year + 1900;
    fp_hdr.pdate.month = tms->tm_mon;
    fp_hdr.pdate.day = tms->tm_mday;
    fp_hdr.pdate.hour = tms->tm_hour;
    fp_hdr.pdate.min = tms->tm_min;
    fp_hdr.pdate.sec = tms->tm_sec;
    fp_hdr.baud = 0;
    fp_hdr.PacketType = 2;
    fp_hdr.origNet = mailproc->baseAddr.net;
    fp_hdr.destNet = mailproc->currAddress->net;
    fp_hdr.prodCode = 0; /*0x17a; */
    fp_hdr.serialNo = 0;
    memset(fp_hdr.password, '\0', sizeof fp_hdr.password);
    fp_hdr.origZone = mailproc->baseAddr.zone;
    fp_hdr.destZone = mailproc->currAddress->zone;
    fp_hdr.origPoint = mailproc->baseAddr.point;
    fp_hdr.destPoint = mailproc->currAddress->point;
    fp_hdr.capability = 0x0001;
    fp_hdr.Revision = 0;
    memcpy(fp_hdr.prodData, "XPKT", sizeof fp_hdr.prodData);
    cvt_to_raw_packet_hdr(&fp_hdr, &raw_p);
    efileWrite(mailproc->currDest, &raw_p, 1, sizeof raw_p);
    return;
}




/*******************************/
/* start of sendechomail stuff  */
/* purely echomail stuff       */
/* purely for incoming packets */
/*******************************/

static void mailprocSendEchomail(MAILPROC *mailproc)
{
    if (mailproc->pkt.startPtr == NULL)
    {
        pktBegin(&mailproc->pkt);
    }
    mailproc->fm_hdr.origNet = mailproc->baseAddr.net;
    mailproc->fm_hdr.origNode = mailproc->baseAddr.node;
    mailproc->fm_hdr.destNet =  mailproc->currAddress->net;
    mailproc->fm_hdr.destNode =  mailproc->currAddress->node;
    efileWrite(mailproc->currDest, "\x02\x00", 1, 2);
    cvt_to_raw_message_hdr(&mailproc->fm_hdr, &mailproc->raw_m);
    efileWrite(mailproc->currDest, &mailproc->raw_m, 1, sizeof mailproc->raw_m);
    efileWrite(mailproc->currDest,
               mailproc->pkt.date,
               1, 
               strlen((char *)mailproc->pkt.date) + 1);
    efileWrite(mailproc->currDest,
               mailproc->pkt.to,
               1, 
               strlen((char *)mailproc->pkt.to) + 1);
    efileWrite(mailproc->currDest,
               mailproc->pkt.from, 
               1, 
               strlen((char *)mailproc->pkt.from) + 1);
    efileWrite(mailproc->currDest,
               mailproc->pkt.subject, 
               1, 
               strlen((char *)mailproc->pkt.subject) + 1);
    mailprocWriteEchoText(mailproc);
    mailproc->totalmsgsOut++;
    return;
}  

static void mailprocWriteEchoText(MAILPROC *mailproc)
{
    long x;
    long numFull;
    long numHalf;
    
    if (mailproc->pkt.endPtr == NULL)
    {
        efileWrite(mailproc->currDest,
                   mailproc->pkt.startValT,
                   1,
                   (size_t)(mailproc->pkt.endBuf - mailproc->pkt.startValT));
        numFull = mailproc->pkt.numHalves / 2;
        numHalf = mailproc->pkt.numHalves % 2;
        for (x = 0; x < numFull; x++)
        {
            pktFull(&mailproc->pkt);
            efileWrite(mailproc->currDest,
                       mailproc->pkt.startBuf,
                       1,
                       (size_t)(mailproc->pkt.endBuf - mailproc->pkt.startBuf));
        }
        if (numHalf == 1)
        {
            pktHalf(&mailproc->pkt);
        }
    }
    mailprocWriteEchoText2(mailproc);
    return;
}

static void mailprocWriteEchoText2(MAILPROC *mailproc)
{
    unsigned char *firstS;
    unsigned char *firstE;
    unsigned char *lastS;
    unsigned char *lastE;
    unsigned char *upto;
    int seenbyFirst = 1;
    int seenbyOnly = 1;

    firstS = mailproc->startSeenby;
    firstE = mailproc->endSeenby;
    lastS = mailproc->startPath;
    lastE = mailproc->endPath;
    if (mailproc->numPath > 0)
    {
        seenbyOnly = 0;
        if (mailproc->startPath < mailproc->startSeenby)
        {
            firstS = mailproc->startPath;
            firstE = mailproc->endPath;
            lastS = mailproc->startSeenby;
            lastE = mailproc->endSeenby;
            seenbyFirst = 0;
        }
    }
    efileWrite(mailproc->currDest,
               mailproc->pkt.startValT,
               1,
               (size_t)(firstS - mailproc->pkt.startValT));
    if (seenbyFirst)
    {
        mailprocWriteSeenby(mailproc);
    }
    else
    {
        mailprocWritePath(mailproc);
    }
    upto = firstE;
    if (!seenbyOnly)
    {
        if (upto != lastS)
        {
            efileWrite(mailproc->currDest,
                       upto,
                       1,
                       (size_t)(lastS - upto));
        }
        if (!seenbyFirst)
        {
            mailprocWriteSeenby(mailproc);
        }
        else
        {
            mailprocWritePath(mailproc);
        }
        upto = lastE;
    }
    else
    {
        mailprocWritePath(mailproc);
    }
    efileWrite(mailproc->currDest,
               upto,
               1,
               (size_t)(mailproc->pkt.endPtr - upto) + 1); /* included NUL */
    return;
}

static void mailprocWriteSeenby(MAILPROC *mailproc)
{
    while (seenbyExtract(&mailproc->seenby,
                         mailproc->newSeenby,
                         sizeof mailproc->newSeenby) == 0)
    {
        if (mailproc->debug)
        {
            printf("new seenby is %s\n", mailproc->newSeenby);
        }
        efileWrite(mailproc->currDest, "SEEN-BY: ", 1, 9);
        efileWrite(mailproc->currDest,
                   mailproc->newSeenby, 
                   1, 
                   strlen((char *)mailproc->newSeenby));
        efileWrite(mailproc->currDest, "\x0d", 1, 1);
    }
    return;
}

static void mailprocWritePath(MAILPROC *mailproc)
{
    while (seenbyExtract(&mailproc->path,
                         mailproc->newPath,
                         sizeof mailproc->newPath) == 0)
    {
        if (mailproc->debug)
        {
            printf("new path is %s\n", mailproc->newPath);
        }
        efileWrite(mailproc->currDest, "\x01PATH: ", 1, 7);
        efileWrite(mailproc->currDest,
                   mailproc->newPath, 
                   1, 
                   strlen((char *)mailproc->newPath));
        efileWrite(mailproc->currDest, "\x0d", 1, 1);
    }
    return;
}

/****************************/
/* end of sendechomail stuff */
/****************************/




/******************************/
/* start of sendnetmail stuff */
/******************************/

static void mailprocSendNetmail(MAILPROC *mailproc,
                                void (*handler)(MAILPROC *mailproc))
{
    int y;
    char *mode;
    
    for (y = 0; y < mailproc->destCount; y++)
    {
        if (fidoAddressCompare(&mailproc->routeAddr,
                               &mailproc->dests[y].addr))
        {
            mailproc->currAddress =
                &mailproc->dests[y].addr;
            if (!mailproc->dests[y].packetOpened)
            {
                mailprocNewPacket(mailproc);
                strcpy(mailproc->dests[y].packetName, 
                       mailproc->newPacket);
                if (exists(mailproc->newPacket))
                {
                    mode = "r+b";
                }
                else
                {
                    mode = "wb";
                }       
                efileOpen(&mailproc->dests[y].packetFile,
                          mailproc->newPacket,
                          mode);
                if (ALLOK)
                {
                    mailproc->dests[y].packetOpened = 1;
                    mailproc->currDest = &mailproc->dests[y].packetFile;
                    if (strcmp(mode, "wb") == 0)
                    {
                        mailprocPacketHeader(mailproc);
                    }
                    else
                    {
                        efileSeek(&mailproc->dests[y].packetFile,
                                  -2L,
                                  SEEK_END);
                    }              
                }
            }
            if (mailproc->dests[y].packetOpened)
            {
                mailproc->currDest = &mailproc->dests[y].packetFile;
                handler(mailproc);
            }
        }
    }
    return;
}

static void mailprocSendNetmail2(MAILPROC *mailproc)
{
    if (mailproc->pkt.startPtr == NULL)
    {
        pktBegin(&mailproc->pkt);
    }
    efileWrite(mailproc->currDest, "\x02\x00", 1, 2);
    cvt_to_raw_message_hdr(&mailproc->fm_hdr, &mailproc->raw_m);
    efileWrite(mailproc->currDest, &mailproc->raw_m, 1, sizeof mailproc->raw_m);
    efileWrite(mailproc->currDest,
               mailproc->pkt.date,
               1, 
               strlen((char *)mailproc->pkt.date) + 1);
    efileWrite(mailproc->currDest,
               mailproc->pkt.to,
               1, 
               strlen((char *)mailproc->pkt.to) + 1);
    efileWrite(mailproc->currDest,
               mailproc->pkt.from, 
               1, 
               strlen((char *)mailproc->pkt.from) + 1);
    efileWrite(mailproc->currDest,
               mailproc->pkt.subject, 
               1, 
               strlen((char *)mailproc->pkt.subject) + 1);
    if (mailproc->remapped)
    {
        mailprocWriteIntl(mailproc);
    }
    mailprocWriteNetmailText(mailproc);
    return;
}  

static void mailprocWriteNetmailText(MAILPROC *mailproc)
{
    long x;
    long numFull;
    long numHalf;
    
    if (mailproc->pkt.endPtr == NULL)
    {
        efileWrite(mailproc->currDest,
                   mailproc->pkt.startValT,
                   1,
                   (size_t)(mailproc->pkt.endBuf - mailproc->pkt.startValT));
        numFull = mailproc->pkt.numHalves / 2;
        numHalf = mailproc->pkt.numHalves % 2;
        for (x = 0; x < numFull; x++)
        {
            pktFull(&mailproc->pkt);
            efileWrite(mailproc->currDest,
                       mailproc->pkt.startBuf,
                       1,
                       (size_t)(mailproc->pkt.endBuf - mailproc->pkt.startBuf));
        }
        if (numHalf == 1)
        {
            pktHalf(&mailproc->pkt);
        }
    }
    efileWrite(mailproc->currDest,
               mailproc->pkt.startValT,
               1,
               (size_t)(mailproc->pkt.endPtr - mailproc->pkt.startValT));
    mailprocWriteVia(mailproc);
    efileWrite(mailproc->currDest,
               "\x00",
               1,
               1);
    return;
}

static void mailprocWriteIntl(MAILPROC *mailproc)
{    
    /*efilePrintf(mailproc->currDest,
                "\x01" "INTL %d:%d/%d %d:%d/%d\r",
                mailproc->msgDest.zone,
                mailproc->msgDest.net,
                mailproc->msgDest.node,
                mailproc->msgOrigin.zone,
                mailproc->msgOrigin.net,
                mailproc->msgOrigin.node);
    if (mailproc->msgOrigin.point != 0)
    {
        efilePrintf(mailproc->currDest, 
                    "\x01" "FMPT %d\r", 
                    mailproc->msgOrigin.point);
    } */
    if (mailproc->recipAddr.point != 0)
    {
        efilePrintf(mailproc->currDest,
                    "\x01" "TOPT %d\r",
                    mailproc->recipAddr.point);
    }
    return;
}

static void mailprocWriteVia(MAILPROC *mailproc)
{
    char buf[50];
    time_t tt;
    struct tm *tms;
    
    time(&tt);
    tms = localtime(&tt);
    sprintf(buf, 
            "%04d-%02d-%02d %02d:%02d:%02d",
            tms->tm_year + 1900,
            tms->tm_mon + 1,
            tms->tm_mday,
            tms->tm_hour,
            tms->tm_min,
            tms->tm_sec);
    efilePrintf(mailproc->currDest,
                "\x01" "Via Tobruk 1.00, %d:%d/%d.%d %s\r",
                mailproc->baseAddr.zone,
                mailproc->baseAddr.net,
                mailproc->baseAddr.node,
                mailproc->baseAddr.point,
                buf);
    return;
}


/****************************/
/* end of sendnetmail stuff */
/****************************/






static void mailprocCurrentArea(MAILPROC *mailproc)
{
    int x;
    
    if (memcmp(mailproc->conference, 
               mailproc->lastArea, 
               mailproc->lastAreaLen) == 0)
    {
        x = mailproc->lastAreaNum;
    }
    else
    {
        for (x=0; x < mailproc->numAreas; x++)
        {
            if (strcmp((char *)mailproc->conference, 
                       mailproc->areas[x].areaName) == 0)
            {
                strcpy(mailproc->lastArea, (char *)mailproc->conference);
                mailproc->lastAreaLen = 
                    strlen((char *)mailproc->conference) + 1;
                mailproc->lastAreaNum = x;
                break;
            }
        }
    }
    if (x == mailproc->numAreas)
    {
        errorSet(MAILPROC_UNKNOWN_AREA, mailproc->conference);
    }
    else
    {
        if (mailproc->debug)
        {
            printf("found area number %d\n", x);
        }
        mailproc->currAreaDef = &mailproc->areas[x];
    }
    return;
}
    
static void mailprocGetDests(MAILPROC *mailproc)
{
    int x;
    int ret;
    
    mailprocMessageBase(mailproc);
    for (x = 0; x < mailproc->currAreaDef->numNodes; x++)
    {
        if (mailproc->currAreaDef->outList[x].fa.point == 0)
        {
            ret = seenbyAddNode(&mailproc->seenby, 
                                &mailproc->currAreaDef->outList[x].fa);
            mailproc->currAreaDef->outList[x].process = (ret == 0);
        }
        else
        {
            ret = fidoAddressCompare(&mailproc->currAreaDef->outList[x].fa,
                                     &mailproc->pktOrigin);
            mailproc->currAreaDef->outList[x].process = (ret == 0);
            if (mailproc->debug)
            {
                printf("going to process point %d\n", 
                       mailproc->currAreaDef->outList[x].fa.point);
            }
        }
    }
    seenbyAddNode(&mailproc->seenby, &mailproc->baseAddr);
    return;
}

void mailprocNumAreas(MAILPROC *mailproc, int numAreas)
{
    size_t amount;
    
    mailproc->numAreas = numAreas;
    amount = numAreas * sizeof *mailproc->areas;
    mailproc->areas = emalloc(amount);
    mailproc->areaCount = -1;
    return;
}

void mailprocAddArea(MAILPROC *mailproc, 
                     char *areaName,
                     char *messageBase, 
                     int numNodes,
                     int passThrough,
                     int mbType,
                     fidoAddress *primary)
{
    mailproc->areaCount++;
    strcpy(mailproc->areas[mailproc->areaCount].areaName, areaName);
    mailproc->areas[mailproc->areaCount].numNodes = numNodes;
    mailproc->areas[mailproc->areaCount].currNode = -1;
    strcpy(mailproc->areas[mailproc->areaCount].messageBase, messageBase);
    mailproc->areas[mailproc->areaCount].passThrough = passThrough;
    mailproc->areas[mailproc->areaCount].mbType = mbType;
    mailproc->areas[mailproc->areaCount].primary = *primary;
    if (numNodes != 0)
    {
        mailproc->areas[mailproc->areaCount].outList =
            emalloc(numNodes * sizeof(mailprocOutNode));
    }
    return;
}

void mailprocNumRemaps(MAILPROC *mailproc, int numRemaps)
{
    size_t amount;
    
    mailproc->numRemaps = numRemaps;
    amount = numRemaps * sizeof *mailproc->remaps;
    mailproc->remaps = emalloc(amount);
    mailproc->remapCount = -1;
    return;
}

void mailprocAddRemap(MAILPROC *mailproc, 
                      char *name,
                      fidoAddress *addr)
{
    mailproc->remapCount++;
    strcpy(mailproc->remaps[mailproc->remapCount].name, name);
    mailproc->remaps[mailproc->remapCount].addr = *addr;
    return;
}

void mailprocAddNodeOnly(MAILPROC *mailproc, FADDR *fidoAddr)
{
    int x;
    fidoAddress fa;
    int found = 0;

    faddrGetaddr(fidoAddr, &fa);
    for (x = 0; x < mailproc->destCount; x++)
    {
        if (fidoAddressCompare(&mailproc->dests[x].addr, &fa))
        {
            found = 1;
            break;
        }
    }
    if (!found)
    {
        if ((mailproc->destCount + 1) == MAILPROC_MAX_DESTS)
        {
            errorSet(MAILPROC_TOO_MANY_DESTS, MAILPROC_MAX_DESTS);
        }
        else
        {
            mailproc->dests[mailproc->destCount].addr = fa;
            mailproc->dests[mailproc->destCount].packetOpened = 0;
            mailproc->destCount++;
        }
    }
    return;
}

void mailprocAddNode(MAILPROC *mailproc, FADDR *fidoAddr)
{
    int x;
    fidoAddress fa;

    x = ++mailproc->areas[mailproc->areaCount].currNode;
    faddrGetaddr(fidoAddr, &fa);
    mailproc->areas[mailproc->areaCount].outList[x].fa = fa;
    mailprocAddNodeOnly(mailproc, fidoAddr);
    return;
}

/* return 1 if addresses compare equal */

static int fidoAddressCompare(fidoAddress *fa1, fidoAddress *fa2)
{
    int ret = 0;
    
/*    printf("comparing %d:%d/%d and %d:%d/%d\n",
           fa1->zone, fa1->net, fa1->node,
           fa2->zone, fa2->net, fa2->node); */
    if ((fa1->node == fa2->node)
        && (fa1->point == fa2->point)
        && (fa1->net == fa2->net)
        && (fa1->zone == fa2->zone)
        && (strcmp(fa1->domain, fa2->domain) == 0))
    {
        ret = 1;
    }
    return (ret);
}

static int miniAddressCompare(fidoAddress *fa1, fidoAddress *fa2)
{
    int ret = 0;
    
    if ((fa1->node == fa2->node)
        && (fa1->net == fa2->net))
    {
        ret = 1;
    }
    return (ret);
}

static void mailprocNewPacket(MAILPROC *mailproc)
{
#ifndef BINKLEYSTYLE
    time_t tt;
    struct tm *tms;
    unsigned long val;
#endif    
    
#ifdef BINKLEYSTYLE
    if (mailproc->currAddress->point == 0)
    {
        sprintf(mailproc->newPacket,
                "%s\\%.4x%.4x.out",
                mailproc->outbound,
                mailproc->currAddress->net,
                mailproc->currAddress->node);
    }
    else
    {
        sprintf(mailproc->newPacket,
                "%s\\%.4x%.4x.pnt\\%.8x.out",            
                mailproc->outbound,
                mailproc->currAddress->net,
                mailproc->currAddress->node,
                mailproc->currAddress->point);
    }
#else    
    time(&tt);
    tms = localtime(&tt);
    val = tms->tm_yday;
    val = val * 24 + tms->tm_hour;
    val = val * 60 + tms->tm_min;
    val = val * 60 + tms->tm_sec;
    val = (val << 8) | mailproc->sequence;
    mailproc->sequence++;
    if (mailproc->sequence == 256)
    {
        mailproc->sequence = 0;
    }
    sprintf(mailproc->newPacket, 
            "%s%.8lX.PKT", 
            mailproc->outbound,
            (val & 0xFFFFFFFFUL));
#endif    
    return;
}

int cvt_from_raw_message_hdr(const raw_fido_msg_hdr *raw_m,
    fido_msg_hdr *fm_hdr)
{
  fm_hdr->origNode = (raw_m->origNode[1] << 8) | raw_m->origNode[0];
  fm_hdr->destNode = (raw_m->destNode[1] << 8) | raw_m->destNode[0];
  fm_hdr->origNet = (raw_m->origNet[1] << 8) | raw_m->origNet[0];
  fm_hdr->destNet = (raw_m->destNet[1] << 8) | raw_m->destNet[0];
  fm_hdr->Attribute = (raw_m->Attribute[1] << 8) | raw_m->Attribute[0];
  fm_hdr->cost = (raw_m->cost[1] << 8) | raw_m->cost[0];
  return (0);
}

int cvt_to_raw_message_hdr(const fido_msg_hdr *fm_hdr,
                           raw_fido_msg_hdr *raw_m)
{
    raw_m->origNode[0] = (unsigned char)(fm_hdr->origNode & 0xff);
    raw_m->origNode[1] = (unsigned char)((fm_hdr->origNode >> 8) & 0xff);
    raw_m->destNode[0] = (unsigned char)(fm_hdr->destNode & 0xff);
    raw_m->destNode[1] = (unsigned char)((fm_hdr->destNode >> 8) & 0xff);
    raw_m->origNet[0] = (unsigned char)(fm_hdr->origNet & 0xff);
    raw_m->origNet[1] = (unsigned char)((fm_hdr->origNet >> 8) & 0xff);
    raw_m->destNet[0] = (unsigned char)(fm_hdr->destNet & 0xff);
    raw_m->destNet[1] = (unsigned char)((fm_hdr->destNet >> 8) & 0xff);
    raw_m->Attribute[0] = (unsigned char)(fm_hdr->Attribute & 0xff);
    raw_m->Attribute[1] = (unsigned char)((fm_hdr->Attribute >> 8) & 0xff);
    raw_m->cost[0] = (unsigned char)(fm_hdr->cost & 0xff);
    raw_m->cost[1] = (unsigned char)((fm_hdr->cost >> 8) & 0xff);
    return (0);
}

int cvt_from_raw_packet_hdr(const raw_fido_packet_hdr *raw_p,
    fido_packet_hdr *fp_hdr)
{
  memset(fp_hdr, '\0', sizeof *fp_hdr);
  fp_hdr->origNode = (raw_p->origNode[1] << 8) | raw_p->origNode[0];
  fp_hdr->destNode = (raw_p->destNode[1] << 8) | raw_p->destNode[0];
  fp_hdr->pdate.year = (raw_p->year[1] << 8) | raw_p->year[0];
  fp_hdr->pdate.month = (raw_p->month[1] << 8) | raw_p->month[0];
  fp_hdr->pdate.day = (raw_p->day[1] << 8) | raw_p->day[0];
  fp_hdr->pdate.hour = (raw_p->hour[1] << 8) | raw_p->hour[0];
  fp_hdr->pdate.min = (raw_p->minute[1] << 8) | raw_p->minute[0];
  fp_hdr->pdate.sec = (raw_p->second[1] << 8) | raw_p->second[0];
  fp_hdr->baud = (raw_p->baud[1] << 8) | raw_p->baud[0];
  fp_hdr->PacketType = (raw_p->PacketType[1] << 8) | raw_p->PacketType[0];
  fp_hdr->origNet = (raw_p->origNet[1] << 8) | raw_p->origNet[0];
  fp_hdr->destNet = (raw_p->destNet[1] << 8) | raw_p->destNet[0];
  fp_hdr->prodCode = raw_p->prodCode;
  fp_hdr->serialNo = raw_p->serialNo;
  memcpy(fp_hdr->password,raw_p->password,sizeof fp_hdr->password);
  fp_hdr->origZone = (raw_p->origZone[1] << 8) | raw_p->origZone[0];
  fp_hdr->destZone = (raw_p->destZone[1] << 8) | raw_p->destZone[0];
  if ((raw_p->CapabilWord[0] != raw_p->CWvalidationCopy[1])
      || (raw_p->CapabilWord[1] != raw_p->CWvalidationCopy[0]))
  {
      fp_hdr->capability = 0;
  }
  else
  {
      fp_hdr->capability = (raw_p->CapabilWord[1] << 8)
                           | raw_p->CapabilWord[0];
      fp_hdr->capability &= 0x7fffU;
  }
  if ((fp_hdr->capability & 0x0001U) != 0)
  {
      fp_hdr->origPoint = (raw_p->origPoint[1] << 8) | raw_p->origPoint[0];
      fp_hdr->destPoint = (raw_p->destPoint[1] << 8) | raw_p->destPoint[0];
      fp_hdr->prodCode = (raw_p->prodCodeH << 8) | fp_hdr->prodCode;
      fp_hdr->Revision = raw_p->Revision;
      memcpy(fp_hdr->prodData, raw_p->prodData, sizeof fp_hdr->prodData);
      if ((fp_hdr->origPoint != 0) && (fp_hdr->origNet == 0xffff))
      {
          fp_hdr->origNet = (raw_p->auxNet[1] << 8) | raw_p->auxNet[0];
      }
  }
  return (0);
}

int cvt_to_raw_packet_hdr(const fido_packet_hdr *fp_hdr,
    raw_fido_packet_hdr *raw_p)
{
    memset(raw_p, '\0', sizeof *raw_p);
    raw_p->origNode[0] = (unsigned char)(fp_hdr->origNode & 0xff);
    raw_p->origNode[1] = (unsigned char)((fp_hdr->origNode >> 8) & 0xff);
    raw_p->destNode[0] = (unsigned char)(fp_hdr->destNode & 0xff);
    raw_p->destNode[1] = (unsigned char)((fp_hdr->destNode >> 8) & 0xff);
    raw_p->year[0] = (unsigned char)(fp_hdr->pdate.year & 0xff);
    raw_p->year[1] = (unsigned char)((fp_hdr->pdate.year >> 8) & 0xff);
    raw_p->month[0] = (unsigned char)(fp_hdr->pdate.month & 0xff);
    raw_p->month[1] = (unsigned char)((fp_hdr->pdate.month >> 8) & 0xff);
    raw_p->day[0] = (unsigned char)(fp_hdr->pdate.day & 0xff);
    raw_p->day[1] = (unsigned char)((fp_hdr->pdate.day >> 8) & 0xff);
    raw_p->hour[0] = (unsigned char)(fp_hdr->pdate.hour & 0xff);
    raw_p->hour[1] = (unsigned char)((fp_hdr->pdate.hour >> 8) & 0xff);
    raw_p->minute[0] = (unsigned char)(fp_hdr->pdate.min & 0xff);
    raw_p->minute[1] = (unsigned char)((fp_hdr->pdate.min >> 8) & 0xff);
    raw_p->second[0] = (unsigned char)(fp_hdr->pdate.sec & 0xff);
    raw_p->second[1] = (unsigned char)((fp_hdr->pdate.sec >> 8) & 0xff);
    raw_p->baud[0] = (unsigned char)(fp_hdr->baud & 0xff);
    raw_p->baud[1] = (unsigned char)((fp_hdr->baud >> 8) & 0xff);
    raw_p->PacketType[0] = (unsigned char)(fp_hdr->PacketType & 0xff);
    raw_p->PacketType[1] = (unsigned char)((fp_hdr->PacketType >> 8) & 0xff);
    raw_p->origNet[0] = (unsigned char)(fp_hdr->origNet & 0xff);
    raw_p->origNet[1] = (unsigned char)((fp_hdr->origNet >> 8) & 0xff);
    raw_p->destNet[0] = (unsigned char)(fp_hdr->destNet & 0xff);
    raw_p->destNet[1] = (unsigned char)((fp_hdr->destNet >> 8) & 0xff);
    raw_p->prodCode = (unsigned char)(fp_hdr->prodCode & 0xff);
    raw_p->serialNo = (unsigned char)(fp_hdr->serialNo & 0xff);
    memcpy(raw_p->password, fp_hdr->password, sizeof raw_p->password);
    raw_p->origZone[0] = (unsigned char)(fp_hdr->origZone & 0xff);
    raw_p->origZone[1] = (unsigned char)((fp_hdr->origZone >> 8) & 0xff);
    raw_p->destZone[0] = (unsigned char)(fp_hdr->destZone & 0xff);
    raw_p->destZone[1] = (unsigned char)((fp_hdr->destZone >> 8) & 0xff);
    raw_p->auxNet[0] = (unsigned char)raw_p->destNet[0];
    raw_p->auxNet[1] = (unsigned char)raw_p->destNet[1];
    raw_p->CWvalidationCopy[0] = (unsigned char)
         ((fp_hdr->capability >> 8) & 0xff);
    raw_p->CWvalidationCopy[1] = (unsigned char)(fp_hdr->capability & 0xff);
    raw_p->CapabilWord[0] = raw_p->CWvalidationCopy[1];
    raw_p->CapabilWord[1] = raw_p->CWvalidationCopy[0];
    raw_p->prodCodeH = (unsigned char)((fp_hdr->prodCode >> 8) & 0xff);
    raw_p->origZone2[0] = raw_p->origZone[0];
    raw_p->origZone2[1] = raw_p->origZone[1];
    raw_p->destZone2[0] = raw_p->destZone[0];
    raw_p->destZone2[1] = raw_p->destZone[1];
    raw_p->Revision = (unsigned char)fp_hdr->Revision;
    raw_p->origPoint[0] = (unsigned char)(fp_hdr->origPoint & 0xff);
    raw_p->origPoint[1] = (unsigned char)((fp_hdr->origPoint >> 8) & 0xff);
    raw_p->destPoint[0] = (unsigned char)(fp_hdr->destPoint & 0xff);
    raw_p->destPoint[1] = (unsigned char)((fp_hdr->destPoint >> 8) & 0xff);
    memcpy(raw_p->prodData, fp_hdr->prodData, sizeof raw_p->prodData);
    return (0);
}    

static void mailprocClosePackets(MAILPROC *mailproc)
{
    int x;
    static char junk[2] = { '\0', '\0' };
    
    for (x = 0; x < mailproc->destCount; x++)
    {
        if (mailproc->dests[x].packetOpened)
        {
            mailproc->dests[x].packetOpened = 0;
            efileWrite(&mailproc->dests[x].packetFile, junk, 1, sizeof junk);
            efileClose(&mailproc->dests[x].packetFile);
        }
    }
    return;
}

long mailprocTotalTime(MAILPROC *mailproc)
{
    time_t endTime;
    long numsecs;
    
    time(&endTime);
    numsecs = (long)difftime(endTime, mailproc->startTime);
    return (numsecs);
}

void mailprocNetmail(MAILPROC *mailproc, char *base, int mbType)
{
    strcpy(mailproc->netarea.areaName, "NETMAIL");
    strcpy(mailproc->netarea.messageBase, base);
    mailproc->netarea.numNodes = 0;
    mailproc->netarea.currNode = 0;
    mailproc->netarea.passThrough = 0;
    mailproc->netarea.outList = NULL;
    mailproc->netarea.mbType = mbType;
#ifdef USE_MSGAPI
    mailproc->netarea.mh = NULL;
#endif          
    return;
}
    
/* parse the origin line */    
    
static void mailprocMsgOrigin(MAILPROC *mailproc)
{
    FADDR faddr;
    char *pos1;
    char *pos2;
    char *pos3;
    size_t lenmove;
    char buf[200];

    faddrDefaults(&faddr);
    faddrInit(&faddr);
    pos2 = strrchr((char *)mailproc->origin, ')');
    if ((pos2 != NULL) && (pos2 != (char *)mailproc->origin))
    {
        pos2--;
        pos1 = pos2;
        while ((*pos1 != '(') && (pos1 > (char *)mailproc->origin))
        {
            pos1--;
        }
        if (pos1 != (char *)mailproc->origin)
        {
            pos1++;
            pos3 = strchr(pos1, ' ');
            if ((pos3 != NULL) && (pos3 < pos2))
            {
                pos1 = pos3;
            }
            while (*pos1 == 0x20)
            {
                pos1++;
            }
            lenmove = (size_t)(pos2 - pos1) + 1;
            if (lenmove > (sizeof buf - 1))
            {
                lenmove = sizeof buf - 1;
            }
            memcpy(buf, pos1, lenmove);
            buf[lenmove] = '\0';
            faddrCreate(&faddr, buf);
            faddrGetaddr(&faddr, &mailproc->authorAddr);
        }
    }
    faddrTerm(&faddr);
    return;
}

#ifdef USE_MBASE        
static void mailprocMsgWrite(MAILPROC *mailproc, GMD *msg)
{
    unsigned char *text;
    size_t textLen;
    long totLen;
    int fin; /* finished writing text */

    text = mailproc->pkt.text;
    textLen = mailproc->pkt.len1;
    totLen = mailproc->pkt.len1;
    fin = mailproc->pkt.endPtr != NULL;
    mbaseInsertMsg(&mailproc->mbase, msg, totLen, text, textLen, fin);
    if (mailproc->pkt.endPtr == NULL)
    {
#if 0    
            offset = textLen;
            textLen = mailproc->pkt.len1;
            numFull = mailproc->pkt.numHalves / 2;
            numHalf = mailproc->pkt.numHalves % 2;
            for (y = 0; (y < numFull) && (ret == 0); x++)
            {
                pktFull(&mailproc->pkt);
                textLen = mailproc->pkt.len1;
                ret = mbaseUpdateText(&mailproc->mbase, 
                                      gth, 
                                      offset,
                                      text,
                                      textLen);
                offset += textLen;
            }
            if ((numHalf == 1) && (ret == 0))
            {
                pktHalf(&mailproc->pkt);
                textLen = mailproc->pkt.len1;
                ret = mbaseUpdateText(&mailproc->mbase, 
                                      gth, 
                                      offset,
                                      text, 
                                      textLen);
                offset += textLen;
            }
#endif            
    }
    return;
}
#endif
    
#ifdef USE_MSGAPI
static void mailprocMsgWrite(MAILPROC *mailproc, MSGH *msgh, XMSG *msg)
{
    unsigned char *ctext;
    dword ctextLen;
    unsigned char *text;
    unsigned char *p;
    char *q;
    char *r;
    dword textLen;
    dword totLen;
    size_t x;
    long y;
    long numFull;
    long numHalf;
    int req1;
    int ret;

    ctext = mailproc->pkt.text;
    if (mailproc->echoFlag)
    {
        p = (unsigned char *)strchr((char *)ctext, 0x0d);
        if (p != NULL)
        {
            p++;
            if (*p == 0x0a)
            {
                p++;
            }
            ctext = p;
        }
    }
    req1 = 1;
    p = ctext;
    for (x = 0; ((x < mailproc->maxControl) && *p);)
    {
        if (req1 && (*p != 0x01))
        {
            break;
        }
        else if (*p == 0x0d)
        {
            req1 = 1;
            p++;
            if (*p == 0x0a)
            {
                p++;
            }
        }
        else
        {
            req1 = 0;
            mailproc->tempControl[x] = *p;
            p++;
            x++;
        }
    }
    mailproc->tempControl[x] = '\0';
    ctextLen = x;
    text = p;
    q = (char *)mailproc->tempControl;
    r = strstr(q, "\x01" "INTL");
    if (r != NULL)
    {
        q = strchr(r + 1, '\x01');
        if (q != NULL)
        {
            memmove(r, q, strlen(q) + 1);
        }
        else
        {
            *r = 0;
        }
        ctextLen = strlen((char *)mailproc->tempControl);
    }
    if (ctextLen == 0)
    {
        ctext = NULL;
    }
    else
    {
        ctextLen++; /* include terminating NUL in length */
        ctext = mailproc->tempControl;
    }
    if (mailproc->pkt.endPtr != NULL)
    {
        textLen = mailproc->pkt.endPtr - text;
        totLen = textLen;
        ret = MsgWriteMsg(msgh, 0, msg, text, textLen, totLen, ctextLen,
                          ctext);
    }    
    else
    {
        textLen = mailproc->pkt.fendBuf - text;
        totLen = (mailproc->pkt.numHalves + 2) * (long)mailproc->pkt.szhbuf;
        ret = MsgWriteMsg(msgh, 0, msg, text, textLen, totLen, ctextLen,
                          ctext);
        if (ret == 0)
        {
            numFull = mailproc->pkt.numHalves / 2;
            numHalf = mailproc->pkt.numHalves % 2;
            for (y = 0; (y < numFull) && (ret == 0); y++)
            {
                pktFull(&mailproc->pkt);
                ret = MsgWriteMsg(msgh, 1, msg, text, textLen, totLen, 
                                  0, ctext);
            }
            if ((numHalf == 1) && (ret == 0))
            {
                pktHalf(&mailproc->pkt);
                textLen = mailproc->pkt.len1;
                ret = MsgWriteMsg(msgh, 1, msg, text, textLen, totLen, 
                                  0, ctext);
            }
        }                          
    }
    if (ret != 0)
    {
        errorSet(MAILPROC_MSGWRITE_FAIL,
                 msgapierr,
                 (long)totLen, 
                 (long)textLen, 
                 (long)ctextLen);
    }
    return;
}
#endif  





/***************************/
/* start of scanbase stuff */
/***************************/

void mailprocScanBase(MAILPROC *mailproc)
{
#if (!defined(USE_MSGAPI) && !defined(USE_MBASE))
    unused(mailproc);
    return;
#endif    
#ifdef USE_MBASE
    long x;
    long y;
    long numAreas;
    long numMessages;
    GAH gah;
    GAD gad;
    MSET mset;
    GMH gmh;
    
    mailproc->mainbuf[mailproc->mnbufsz] = '\0';
    mailproc->echoFlag = 1;
    mbaseAsetGetCount(&mailproc->mbase, &mailproc->aset, &numAreas);
    for (x = 0; x < numAreas; x++)
    {
        if (x == (numAreas - 1)) /* hack for netmail */
        {
            mailproc->currAreaDef = &mailproc->netarea;
            mailproc->echoFlag = 0;
        }
        else
        {
            mailproc->currAreaDef = &mailproc->areas[(size_t)x];
            mailproc->baseAddr = mailproc->currAreaDef->primary;
        }
        mbaseAsetGetRel(&mailproc->mbase, 
                        &mailproc->aset, 
                        &gah,
                        x);
        mbaseGetAreaData(&mailproc->mbase, &gah, &gad);
        if (!gad.passThrough)
        {
            mbaseGetMessagesInArea(&mailproc->mbase, &gah, &mset);
            mbaseMsetGetCount(&mailproc->mbase, 
                              &mset, 
                              &numMessages);
            for (y = 0; y < numMessages; y++)
            {
                mbaseMsetGetRel(&mailproc->mbase,
                                &mset,
                                &gmh,
                                y);
                mbaseOpenMessage(&mailproc->mbase,
                                 &gmh,
                                 &mailproc->gmd,
                                 MBASE_RW);
                mbaseAccessMessage(&mailproc->mbase,
                                   &mailproc->gmd,
                                   MBASE_READHDR,
                                   NULL,
                                   0,
                                   0,
                                   NULL);
                if (!(mailproc->gmd.attr & ATTR_SENT)
                    && (mailproc->gmd.attr & ATTR_LOCAL))
                {
                    mbaseAccessMessage(&mailproc->mbase,
                                       &mailproc->gmd,
                                       MBASE_READTEXT,
                                       mailproc->mainbuf,
                                       0,
                                       mailproc->mainbufSize,
                                       NULL);
                    mailproc->msglen = strlen((char *)mailproc->mainbuf);
                    mailprocFetchMessage3(mailproc);
                    mailproc->gmd.attr |= ATTR_SENT;
                    mbaseAccessMessage(&mailproc->mbase,
                                       &mailproc->gmd,
                                       MBASE_UPDHDR,
                                       NULL,
                                       0,
                                       0,
                                       NULL);
                }
                mbaseCloseMessage(&mailproc->mbase,
                                  &mailproc->gmd);
            }
        }
    }
    mailproc->baseAddr = mailproc->mainAddr;
    return;
#endif
#ifdef USE_MSGAPI
    int x;
    
    mailproc->mainbuf[mailproc->mnbufsz] = '\0';
    mailproc->echoFlag = 1;
    for (x = 0; x < mailproc->numAreas; x++)
    {
        if (!mailproc->areas[x].passThrough)
        {
            mailproc->currAreaDef = &mailproc->areas[x];
            mailproc->baseAddr = mailproc->currAreaDef->primary;
            mailprocFetchMessage(mailproc);
            mailproc->baseAddr = mailproc->mainAddr;
        }
    }
    if (ALLOK)
    {
        mailproc->echoFlag = 0;
        mailproc->currAreaDef = &mailproc->netarea;
        mailprocFetchMessage(mailproc);
    }
    return;
#endif
}

#ifdef USE_MBASE
static void mailprocFetchMessage3(MAILPROC *mailproc)
{
    mailproc->msglen = strlen((char *)mailproc->mainbuf);
    mailproc->allin = 1;
    if (mailproc->echoFlag)
    {
        mailprocScourEndMsg(mailproc);
        mailprocCalcSeenby(mailproc);
        mailprocCalcPath(mailproc);
        mailprocIterateDests(mailproc, mailprocMsgOut);
    }
    else
    {
        mailprocNetmailMsgSetup(mailproc);
    }
    return;
}
#endif

#ifdef USE_MSGAPI
static void mailprocFetchMessage(MAILPROC *mailproc)
{
    word type;
    
    if (mailproc->currAreaDef->mh == NULL)
    {
        if (mailproc->currAreaDef->mbType == MAILPROC_MB_SQUISH)
        {
            type = MSGTYPE_SQUISH;
        }
        else
        {
            type = MSGTYPE_SDM;
        }
        if (mailproc->echoFlag)
        {
            type |= MSGTYPE_ECHO;
        }
        mailproc->currAreaDef->mh = 
            MsgOpenArea((byte *)mailproc->currAreaDef->messageBase,
                        MSGAREA_CRIFNEC,
                        type);
    }
    if (mailproc->currAreaDef->mh == NULL)
    {
        errorSet(MSG_AREA_OPEN, mailproc->currAreaDef->areaName, msgapierr);
    }
    else
    {
        mailproc->highWater = MsgGetHighWater(mailproc->currAreaDef->mh);
        mailprocFetchMessage2(mailproc);
        if (ALLOK)
        {
            MsgSetHighWater(mailproc->currAreaDef->mh, mailproc->highWater);
        }
        MsgCloseArea(mailproc->currAreaDef->mh);
        mailproc->currAreaDef->mh = NULL;
    }
    return;
}

static void mailprocFetchMessage2(MAILPROC *mailproc)
{
    mailproc->msgh = MsgOpenMsg(mailproc->currAreaDef->mh,
                                MOPEN_RW,
                                mailproc->highWater + 1);
    while (ALLOK && (mailproc->msgh != NULL))
    {
        mailprocFetchMessage3(mailproc);
        MsgCloseMsg(mailproc->msgh);
        if (ALLOK)
        {
            mailproc->highWater++;
            mailproc->msgh = MsgOpenMsg(mailproc->currAreaDef->mh,
                                        MOPEN_RW,
                                        mailproc->highWater + 1);
        }
    }
    return;
}

/*
    get header
    find end of message + origin
    calculate seenbys.
    calculate path
    for all proper destinations
    {
        write out header
        write out area
        write out text
        optionally write out tear + origin
        write out seenby and path
        write out remainder of text.
    }
*/

static void mailprocFetchMessage3(MAILPROC *mailproc)
{
    mailprocGetMsgHeader(mailproc);
    if (((mailproc->xmsg.attr & MSGSCANNED) == 0)
        && ((mailproc->xmsg.attr & MSGLOCAL) != 0))
    {
        mailproc->msglen = strlen((char *)mailproc->mainbuf);
        mailproc->allin = 1;
        if (mailproc->msglen >= mailproc->mnbufsz)
        {
            mailprocFindEndMsg(mailproc);
        }
        if (mailproc->echoFlag)
        {
            mailprocScourEndMsg(mailproc);
            mailprocCalcSeenby(mailproc);
            mailprocCalcPath(mailproc);
            mailprocIterateDests(mailproc, mailprocMsgOut);
        }
        else
        {
            mailprocNetmailMsgSetup(mailproc);
        }
        if (ALLOK)
        {
            mailprocUpdateMsgHeader(mailproc);
        }        
    }
    return;
}
#endif

#if (defined(USE_MBASE) || defined(USE_MSGAPI))
static void mailprocNetmailMsgSetup(MAILPROC *mailproc)
{
    strcpy(mailproc->msgDest.domain, mailproc->baseAddr.domain);
#ifdef USE_MBASE
    if (mailproc->gmd.to_2d.zone == 0)
    {
        mailproc->gmd.to_2d.zone 
            = mailproc->gmd.to_4d.zone
            = mailproc->baseAddr.zone;
    }
    mailproc->msgDest.zone = mailproc->gmd.to_2d.zone;
    mailproc->msgDest.net = mailproc->gmd.to_2d.net;
    mailproc->msgDest.node = mailproc->gmd.to_2d.node;
    mailproc->msgDest.point = mailproc->gmd.to_2d.point;
#endif    
#ifdef USE_MSGAPI    
    mailproc->msgDest.zone = mailproc->xmsg.dest.zone;
    mailproc->msgDest.net = mailproc->xmsg.dest.net;
    mailproc->msgDest.node = mailproc->xmsg.dest.node;
    mailproc->msgDest.point = mailproc->xmsg.dest.point;
#endif    
    
    strcpy(mailproc->msgOrigin.domain, mailproc->baseAddr.domain);
#ifdef USE_MBASE
    mailproc->msgOrigin.zone = mailproc->gmd.from_2d.zone;
    mailproc->msgOrigin.net = mailproc->gmd.from_2d.net;
    mailproc->msgOrigin.node = mailproc->gmd.from_2d.node;
    mailproc->msgOrigin.point = mailproc->gmd.from_2d.point;
#endif    
#ifdef USE_MSGAPI    
    mailproc->msgOrigin.zone = mailproc->xmsg.orig.zone;
    mailproc->msgOrigin.net = mailproc->xmsg.orig.net;
    mailproc->msgOrigin.node = mailproc->xmsg.orig.node;
    mailproc->msgOrigin.point = mailproc->xmsg.orig.point;
#endif    

    mailproc->recipAddr = mailproc->msgDest;
    mailproc->authorAddr = mailproc->msgOrigin;
    
    mailproc->remapped = 0;
    mailproc->routed = 0;
    if (fidoAddressCompare(&mailproc->recipAddr, &mailproc->baseAddr))
    {
#ifdef USE_MBASE
        mailprocRemap(mailproc, (unsigned char *)mailproc->gmd.to);
#endif    
#ifdef USE_MSGAPI    
        mailprocRemap(mailproc, mailproc->xmsg.to);
#endif        
    }
    if (!fidoAddressCompare(&mailproc->recipAddr, &mailproc->baseAddr))
    {
        mailprocRoute(mailproc, &mailproc->msgDest);
        if (mailproc->routed)
        {
            mailprocSendNetmail(mailproc, mailprocMsgOut);
        }
    }
    return;
}
#endif

#ifdef USE_MSGAPI
static void mailprocGetMsgHeader(MAILPROC *mailproc)
{
    mailproc->msgbytes = MsgReadMsg(mailproc->msgh, 
                                    &mailproc->xmsg, 
                                    0, 
                                    mailproc->mnbufsz,
                                    mailproc->mainbuf, 
                                    mailproc->maxControl, 
                                    mailproc->tempControl);
    if (mailproc->msgbytes == (dword)-1)
    {
        errorSet(MSG_READ, msgapierr);
    }
    else    
    {
        mailproc->mainbuf[(size_t)mailproc->msgbytes] = '\0';
    }
    return;
}

static void mailprocUpdateMsgHeader(MAILPROC *mailproc)
{
    int ret;

    if (mailproc->echoFlag || (!mailproc->echoFlag && mailproc->routed))    
    {
        if (mailproc->echoFlag)
        {
            mailproc->xmsg.attr |= MSGSCANNED;
        }
        else
        {
            mailproc->xmsg.attr |= MSGSENT;
            mailproc->xmsg.attr &= ~MSGLOCAL;
        }
        ret = MsgWriteMsg(mailproc->msgh, 
                          1, 
                          &mailproc->xmsg, 
                          NULL,
                          0,
                          0,
                          0,
                          NULL);
        if (ret != 0)
        {
            errorSet(MAILPROC_MSGWRITE_FAIL2, msgapierr);
        }
    }
    return;
}

static void mailprocFindEndMsg(MAILPROC *mailproc)
{
    mailproc->allin = 0;
    mailproc->numhalves = 0;
    while (ALLOK && (mailproc->msglen >= mailproc->mnbufsz))
    {
        mailproc->msgoffs += mailproc->msglen;
        mailproc->msgbytes = MsgReadMsg(mailproc->msgh, 
                                        NULL, 
                                        mailproc->msgoffs, 
                                        mailproc->mnbufsz,
                                        mailproc->mainbuf, 
                                        0,
                                        NULL);
        if (mailproc->msgbytes == (dword)-1)
        {
            errorSet(MSG_READ, msgapierr);
        }
        else    
        {
            mailproc->mainbuf[(size_t)mailproc->msgbytes] = '\0';
            mailproc->msglen = strlen((char *)mailproc->mainbuf);
            mailproc->numhalves += 2;
        }
    }
    if (ALLOK)
    {
        if (mailproc->msglen < mailproc->mnbufhsz)
        {
            mailproc->msgoffs -= mailproc->mnbufhsz;
            mailproc->msgbytes = MsgReadMsg(mailproc->msgh, 
                                            NULL, 
                                            mailproc->msgoffs, 
                                            mailproc->mnbufsz,
                                            mailproc->mainbuf, 
                                            0,
                                            NULL);
            if (mailproc->msgbytes == (dword)-1)
            {
                errorSet(MSG_READ, msgapierr);
            }
            else    
            {
                mailproc->mainbuf[(size_t)mailproc->msgbytes] = '\0';
                mailproc->msglen = strlen((char *)mailproc->mainbuf);
                mailproc->numhalves--;
            }
        }
    }
    return;
}
#endif

#if (defined(USE_MSGAPI) || defined(USE_MBASE))
static void mailprocScourEndMsg(MAILPROC *mailproc)
{
    unsigned char *prev = NULL;
    
    mailproc->msgorig = 
        (unsigned char *)strstr((char *)mailproc->mainbuf, " * Origin:");
    while (mailproc->msgorig != NULL)
    {
        prev = mailproc->msgorig;
        mailproc->msgorig = 
            (unsigned char *)strstr((char *)mailproc->msgorig + 1,
                                    " * Origin:");
    }
    mailproc->msgorig = prev;
    mailproc->startSeenby = NULL;
    if (mailproc->msgorig != NULL)
    {
        mailproc->startSeenby = 
            (unsigned char *)strchr((char *)mailproc->msgorig, 0x0d);
        if (mailproc->startSeenby != NULL)
        {
            mailproc->startSeenby++;
            if (*mailproc->startSeenby == 0x0a)
            {
                mailproc->startSeenby++;
            }
        }
    }
    if (mailproc->startSeenby == NULL)
    {
        mailproc->startSeenby = mailproc->mainbuf + mailproc->msglen;
    }
    return;
}
        
static void mailprocCalcSeenby(MAILPROC *mailproc)
{
    int x;
    int ret;
    
    seenbyReset(&mailproc->seenby);
    for (x = 0; x < mailproc->currAreaDef->numNodes; x++)
    {
        if (mailproc->currAreaDef->outList[x].fa.point == 0)
        {
            ret = seenbyAddNode(&mailproc->seenby, 
                                &mailproc->currAreaDef->outList[x].fa);
            mailproc->currAreaDef->outList[x].process = (ret == 0);
        }
        else
        {
            mailproc->currAreaDef->outList[x].process = 1;
            if (mailproc->debug)
            {
                printf("going to process point %d\n", 
                       mailproc->currAreaDef->outList[x].fa.point);
            }
        }
    }
    seenbyAddNode(&mailproc->seenby, &mailproc->baseAddr);
    seenbySort(&mailproc->seenby);
    return;
}

static void mailprocCalcPath(MAILPROC *mailproc)
{
    seenbyReset(&mailproc->path);
    seenbyAddNode(&mailproc->path, &mailproc->baseAddr);
    return;
}

static void mailprocMsgOut(MAILPROC *mailproc)
{
    mailprocWriteMsgHeader(mailproc);
    if (mailproc->echoFlag)
    {
        mailprocWriteAreaLine(mailproc);
    }
    mailprocWriteKludges(mailproc);
    mailprocWriteTextBody(mailproc);
    return;
}

static void mailprocWriteMsgHeader(MAILPROC *mailproc)
{
    efileWrite(mailproc->currDest, "\x02\x00", 1, 2);
    if (!mailproc->echoFlag)
    {
        mailproc->currAddress = &mailproc->msgDest;
    }
    mailproc->fm_hdr.origNode = mailproc->baseAddr.node;
    mailproc->fm_hdr.destNode = mailproc->currAddress->node;
    mailproc->fm_hdr.origNet = mailproc->baseAddr.net;
    mailproc->fm_hdr.destNet = mailproc->currAddress->net;
    if (mailproc->echoFlag)
    {
        mailproc->fm_hdr.Attribute = 0;
    }
    else
    {
        mailproc->fm_hdr.Attribute = 1;
        if (mailproc->currAddress->zone != mailproc->baseAddr.zone)
        {
            mailproc->fm_hdr.destNode = mailproc->currAddress->zone;
            mailproc->fm_hdr.destNet = mailproc->baseAddr.zone;
        }
    }
    mailproc->fm_hdr.cost = 0;
    cvt_to_raw_message_hdr(&mailproc->fm_hdr, &mailproc->raw_m);    
    efileWrite(mailproc->currDest, 
               &mailproc->raw_m,
               1,
               sizeof mailproc->raw_m);
#ifdef USE_MBASE
    efileWrite(mailproc->currDest,
               mailproc->gmd.create_time,
               1,
               strlen((char *)mailproc->gmd.create_time) + 1);
    efileWrite(mailproc->currDest,
               mailproc->gmd.to,
               1,
               strlen((char *)mailproc->gmd.to) + 1);               
    efileWrite(mailproc->currDest,
               mailproc->gmd.from,
               1,
               strlen((char *)mailproc->gmd.from) + 1);               
    efileWrite(mailproc->currDest,
               mailproc->gmd.subj,
               1,
               strlen((char *)mailproc->gmd.subj) + 1);
#endif               
#ifdef USE_MSGAPI               
    efileWrite(mailproc->currDest,
               mailproc->xmsg.__ftsc_date,
               1,
               strlen((char *)mailproc->xmsg.__ftsc_date) + 1);
    efileWrite(mailproc->currDest,
               mailproc->xmsg.to,
               1,
               strlen((char *)mailproc->xmsg.to) + 1);               
    efileWrite(mailproc->currDest,
               mailproc->xmsg.from,
               1,
               strlen((char *)mailproc->xmsg.from) + 1);               
    efileWrite(mailproc->currDest,
               mailproc->xmsg.subj,
               1,
               strlen((char *)mailproc->xmsg.subj) + 1);
#endif               
    return;
}

static void mailprocWriteAreaLine(MAILPROC *mailproc)
{
    efileWrite(mailproc->currDest,
               "AREA:",
               1,
               5);
    efileWrite(mailproc->currDest,
               mailproc->currAreaDef->areaName,
               1,
               strlen((char *)mailproc->currAreaDef->areaName));
    efileWrite(mailproc->currDest,
               "\x0d",
               1,
               1);               
    return;
}

static void mailprocWriteKludges(MAILPROC *mailproc)
{
    char *p;
    char *q;
    
    p = (char *)mailproc->tempControl;
    while (*p == 0x01)
    {
        q = strchr(p + 1, 0x01);
        if (q == NULL)
        {
            q = p + strlen(p);
        }
        efileWrite(mailproc->currDest,
                   p,
                   1,
                   (size_t)(q - p));
        efileWrite(mailproc->currDest,
                   "\r",
                   1,
                   1);
        p = q;        
    }
    return;
}

static void mailprocWriteTextBody(MAILPROC *mailproc)
{
#ifdef USE_MSGAPI
    int x;
    
    if (!mailproc->allin)
    {    
        mailproc->msgoffs = 0;
        x = mailproc->numhalves;
        while (ALLOK && (x >= 2))
        {
            mailproc->msgbytes = MsgReadMsg(mailproc->msgh, 
                                            NULL, 
                                            mailproc->msgoffs, 
                                            mailproc->mnbufsz,
                                            mailproc->mainbuf, 
                                            0,
                                            NULL);
            if (mailproc->msgbytes == (dword)-1)
            {
                errorSet(MSG_READ, msgapierr);
            }
            else    
            {
                mailproc->mainbuf[(size_t)mailproc->msgbytes] = '\0';
                mailproc->msglen = strlen((char *)mailproc->mainbuf);
                efileWrite(mailproc->currDest,
                           mailproc->mainbuf,
                           1,
                           mailproc->msglen);
                x -= 2;
            }
        }
        if (ALLOK && (x == 1))
        {
            memcpy(mailproc->mainbuf, 
                   mailproc->mainbuf + mailproc->mnbufhsz,
                   mailproc->mnbufhsz);
            mailproc->msgbytes = MsgReadMsg(mailproc->msgh, 
                                            NULL, 
                                            mailproc->msgoffs, 
                                            mailproc->mnbufsz,
                                            mailproc->mainbuf, 
                                            0,
                                            NULL);
            if (mailproc->msgbytes == (dword)-1)
            {
                errorSet(MSG_READ, msgapierr);
            }
            else    
            {
                mailproc->mainbuf[(size_t)mailproc->msgbytes] = '\0';
                mailproc->msglen = strlen((char *)mailproc->mainbuf);
            }
        }
    }
#endif    
    if (!mailproc->echoFlag)
    {
        mailproc->startSeenby = mailproc->mainbuf + mailproc->msglen;
    }
    efileWrite(mailproc->currDest,
               mailproc->mainbuf,
               1,
               (size_t)(mailproc->startSeenby - mailproc->mainbuf));
    if (mailproc->echoFlag)
    {
        mailprocWriteSeenby(mailproc);
        mailprocWritePath(mailproc);    
        efileWrite(mailproc->currDest,
                   mailproc->startSeenby,
                   1,
                   (size_t)((mailproc->mainbuf + mailproc->msglen) 
                            - mailproc->startSeenby));
    }
    else
    {
        mailprocWriteVia(mailproc);
    }
    efileWrite(mailproc->currDest,
               "\x00",
               1,
               1);
    return;
}
#endif

/*************************/
/* end of scanbase stuff */
/*************************/




