/* Written by Paul Edwards */
/* Released to the Public Domain */
/* mbase - generic messagebase routines */

#ifndef MBASE_INCLUDED
#define MBASE_INCLUDED

#include "faddr.h"
#include "area.h"

#define ATTR_SENT 0x0008
#define ATTR_TRANSIT 0x0020
#define ATTR_PRIVATE 0x0001
#define ATTR_LOCAL 0x0100

#define MBASE_RW 0
#define MBASE_READHDR 0
#define MBASE_READTEXT 0
#define MBASE_UPDHDR 1

/* MBASE - message base */

typedef struct
{
    AREADEFN *echolist;
    AREADEFN *netarea;
    long numAreas;
    size_t lastarea;
} MBASE;


/* MSET - set of selected messages */

typedef struct
{
    long count;
    int *list;
} MSET;


/* ASET - set of selected areas */

typedef struct
{
    int x;
} ASET;


/* GMH - generic message handle */

typedef struct
{
    int absNumber;
} GMH;


/* GAD - generic area data */

typedef struct
{
    int passThrough;
} GAD;


/* GAH - generic area handle */

typedef struct
{
    long areanum;
} GAH;


/* FTN mail is based on layers upon layers, vis:

   At the lowest layer, everything is a netmail message going from
   a 2D node to another 2D node.  The netmail is not necessarily
   going directly between the nodes, but may be routed.  The 
   intermediate nodes are expected to be observing that 2D routing.
   
   Interzone netmail is grafted on top of this, via the use of
   zonegates and the INTL kludge.  The sender specifies the 3D
   sending address and 3D receiver address via the INTL line, and
   then proceeds to fill in the 2D address to go from him to the
   zonegate.  The zonegate, on receiving the message, will send
   the message to the other zonegate, who will in turn substitute
   the destination 2D address with one obtained from scanning the
   INTL line.
   
   Echomail is grafted on top of netmail, with an added restriction
   that it must be 2D-addressed mail (ie no INTL).  Usually it is
   also a requirement that the netmail be addressed to someone you
   connect directly with, but this is not a technical requirement.
   The originator of an echomail message specifies his address in
   the Origin line.  He then sends a netmail message with an "AREA"
   line to anyone who wants the echo area and is connected directly
   to him.  The recipient will then in turn generate another netmail
   message to one of his session partners.
   
   Therefore a particular message may have:
   
   from_2d
   to_2d
   from_4d (netmail only - but in echomail, set to orig_4d)
   to_4d (netmail only - but in echomail, set to your address)
   orig_4d (echomail only - but in netmail, set to from_4d)
   
   other addresses of interest are:
   
   who sent you the packet that contained this message (5d address).
   who you exported this message to (5d addresses).
   
*/

/* GMD - generic message data */ 

typedef struct
{
    /* public */
    fidoAddress from_2d;
    fidoAddress to_2d;
    fidoAddress from_4d;
    fidoAddress to_4d;
    fidoAddress orig_4d;
    GAH *gah;
    char from[37];
    char to[37];
    char subj[73];
    unsigned long attr;
    char create_time[20];
    /* private */
    FILE *fu;
} GMD;


/* GUH - generic user handle */

typedef struct
{
    int x;
} GUH;

int mbaseDefaults(MBASE *mbase);
int mbaseInit(MBASE *mbase);
int mbaseTerm(MBASE *mbase);
int mbaseGetAreaList(MBASE *mbase, ASET *aset);
int mbaseAsetQueryByName(MBASE *mbase, ASET *aset, GAH *gah, void *name);
int mbaseAsetDiscard(MBASE *mbase, ASET *aset);
int mbaseAsetGetCount(MBASE *mbase, ASET *aset, long *count);
int mbaseAsetGetRel(MBASE *mbase, ASET *aset, GAH *gah, long relNum);
int mbaseGetAreaData(MBASE *mbase, GAH *gah, GAD *gad);
int mbaseGetMessagesInArea(MBASE *mbase, GAH *gah, MSET *mset);
int mbaseMsetGetCount(MBASE *mbase, MSET *mset, long *count);
int mbaseMsetGetRel(MBASE *mbase, MSET *mset, GMH *gmh, long relnum);
int mbaseMsetDiscard(MBASE *mbase, MSET *mset);
int mbaseSetLastRead(MBASE *mbase, GAH *gah, GUH *guh, GMH *gmh);
int mbaseOpenMessage(MBASE *mbase, GMH *gmh, GMD *gmd, int options);
int mbaseAccessMessage(MBASE *mbase, 
                       GMD *gmd, 
                       int options, 
                       void *text, 
                       size_t start_text, 
                       size_t max_text, 
                       size_t *actual_text);
int mbaseCloseMessage(MBASE *mbase, GMD *gmd);
int mbaseInsertMsg(MBASE *mbase, 
                   GMD *gmd, 
                   long totLen, 
                   unsigned char *text, 
                   size_t textLen, 
                   int fin);
int mbaseProvideAreas(MBASE *mbase, 
                      AREADEFN *areas, 
                      long numAreas, 
                      AREADEFN *netarea);

#endif
