/*********************************************************************/
/*                                                                   */
/*  This Program Written by Paul Edwards, 3:711/934@fidonet.         */
/*  Released to the Public Domain                                    */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  pkt - read Fidonet packets                                       */
/*                                                                   */
/*  Specs:                                                           */
/*                                                                   */
/*  reading a message will set startPtr, pointer to subject, etc.    */
/*  If the message all fits in, endPtr will also be set.             */
/*  Half() will take you half a buffer forward at a time, and        */
/*  endptr will be set as soon as the end is visible.  Note that     */
/*  the first call to Half() will make startPtr invalid (NULL).      */
/*  Full() will go a full buffer at a time.                          */
/*  Begin() takes you back to the beginning of the message.  End()   */
/*  takes you to the end of the message, and will expose at least    */
/*  half a buffer worth of message text.                             */
/*                                                                   */
/*  The date, to, from, subject, text pointers are not changed in    */
/*  any operation except NextMsg() and FirstMsg().                   */
/*                                                                   */
/*  The FirstMsg() and NextMsg() functions will always attempt to    */
/*  expose as much text as possible.                                 */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <string.h>

#include "pkt.h"
#include "slidwin.h"
#include "repofile.h"
#include "estdlib.h"
#include "error.h"

static void pktHalfB(PKT *pkt);
static void pktEndNX(PKT *pkt);
static void pktEndCheck(PKT *pkt);
static void pktSetLoc(PKT *pkt, long offset);
static int pktStdMsg(PKT *pkt);

void pktDefaults(PKT *pkt)
{
    pkt->szbuf = 60000U;
    pkt->buf = NULL;
    pkt->localmem = 1;
    repofileDefaults(&pkt->repofile);
    slidwinDefaults(&pkt->slidwin);
    return;
}

void pktSetSize(PKT *pkt, size_t szbuf)
{
    pkt->szbuf = szbuf;
    return;
}

void pktSetBuf(PKT *pkt, void *buf)
{
    pkt->buf = (unsigned char *)buf;
    pkt->localmem = 0;
    return;
}

void pktInit(PKT *pkt)
{
    if (pkt->localmem)
    {
        pkt->buf = emalloc(pkt->szbuf);
    }
    if (ALLOK)
    {
        pkt->szbuf -= 100; /* allow 100 bytes at end for safety */
        pkt->szhbuf = pkt->szbuf / 2;
        pkt->startBuf = pkt->buf;
        pkt->startPtr = pkt->startBuf;
        pkt->endBuf = pkt->buf + pkt->szbuf;
        *pkt->endBuf = '\0';
        repofileInit(&pkt->repofile);
        if (ALLOK)
        {
            slidwinInit(&pkt->slidwin);
            slidwinSetBuf(&pkt->slidwin, pkt->buf, pkt->szbuf);
        }
        if (!ALLOK)
        {
            repofileTerm(&pkt->repofile);
        }
    }
    if (!ALLOK && pkt->localmem)
    {
        efree(pkt->buf);
    }
    return;
}

void pktTerm(PKT *pkt)
{
    slidwinTerm(&pkt->slidwin);
    repofileTerm(&pkt->repofile);
    if (pkt->localmem)
    {
        efree(pkt->buf);
    }
    return;
}

void pktSetFileBuf(PKT *pkt, void *vbuf, size_t vbufSize)
{
    repofileSetBuf(&pkt->repofile, vbuf, vbufSize);
    return;
}

void pktSetFile(PKT *pkt, char *filename)
{
    repofileSetFile(&pkt->repofile, filename);
    slidwinSetRepo(&pkt->slidwin, repofileGetRepo(&pkt->repofile));
    if (slidwinVal(&pkt->slidwin) < 58)
    {
        errorSet(PKT_INVALID, 1);
    }
    pkt->endOfPkt = 0;
    pkt->currLoc = 0;
    pkt->startPtr = pkt->startBuf;
    pkt->fendBuf = pkt->startBuf + slidwinVal(&pkt->slidwin);
    *pkt->fendBuf = '\0';
    return;
}

void pktFinFile(PKT *pkt)
{
    slidwinFinRepo(&pkt->slidwin);
    return;
}

int pktFirstMsg(PKT *pkt)
{
    /* pktSetLoc(pkt, 58); */
    pkt->startPtr = pkt->startBuf + 58;
    pktStdMsg(pkt);
    return (pkt->endOfPkt);
}

int pktNextMsg(PKT *pkt)
{
    if (pkt->endPtr == NULL)
    {
        pktEndNX(pkt);
    }
    if (ALLOK)
    {
        pkt->remain = (size_t)(pkt->fendBuf - pkt->endPtr);
        pkt->startPtr = pkt->endPtr + 1;
        if (((pkt->remain < 300) && (pkt->fendBuf == pkt->endBuf))
            || (pktStdMsg(pkt) != 0))
        {
            pktSetLoc(pkt, pkt->currLoc + (pkt->startPtr - pkt->startBuf));
            pkt->startPtr = pkt->startBuf;
            pktStdMsg(pkt);
        }
    }
    else
    {
        pkt->endOfPkt = 1;
    }
    return (pkt->endOfPkt);
}

void pktHalf(PKT *pkt)
{
    pkt->startPtr = NULL;
    pkt->startValT = pkt->startBuf;
    pktSetLoc(pkt, pkt->currLoc + pkt->szhbuf);
    pktEndCheck(pkt);
    return;
}

static void pktHalfB(PKT *pkt)
{
    pkt->startPtr = NULL;
    pkt->startValT = pkt->startBuf;
    pktSetLoc(pkt, pkt->currLoc - pkt->szhbuf);
    pktEndCheck(pkt);
    return;
}

void pktFull(PKT *pkt)
{
    pkt->startPtr = NULL;
    pkt->startValT = pkt->startBuf;
    pktSetLoc(pkt, pkt->currLoc + pkt->szbuf);
    pktEndCheck(pkt);
    return;
}

/* Go to end of message without the requirement of exposing at 
least half a buffer's worth */

static void pktEndNX(PKT *pkt)
{
    pkt->numHalves = 2;
    while (pktFull(pkt), ((pkt->endPtr == NULL) && ALLOK))
    {
        pkt->numHalves += 2;
    }
    return;
}

void pktEnd(PKT *pkt)
{
    pktEndNX(pkt);
    if (ALLOK)
    {
        if ((pkt->endPtr - pkt->startBuf) < pkt->szhbuf)
        {
            pktHalfB(pkt);
            pkt->numHalves--;
        }
        pkt->startValT = pkt->startBuf;
    }
    return;
}

static void pktEndCheck(PKT *pkt)
{
    pkt->len1 = strlen((char *)pkt->startBuf);
    if ((pkt->buf + pkt->len1) < pkt->fendBuf)
    {
        pkt->endPtr = pkt->startBuf + pkt->len1;
    }
    return;
}

static void pktSetLoc(PKT *pkt, long offset)
{
    pkt->currLoc = offset;
    slidwinSeek(&pkt->slidwin, pkt->currLoc);
    if (ALLOK)
    {
        pkt->fendBuf = pkt->startBuf + slidwinVal(&pkt->slidwin);
        *pkt->fendBuf = '\0';
    }
    return;
}

static int pktStdMsg(PKT *pkt)
{
    pkt->origLoc = pkt->currLoc;
    pkt->remain = (size_t)(pkt->fendBuf - pkt->startPtr);
    if (pkt->remain < 19)
    {
        pkt->endOfPkt = 1;
    }
    else
    {
        pkt->date = pkt->startPtr + 14;
        pkt->to = pkt->date + strlen((char *)pkt->date) + 1;
        if (pkt->to >= pkt->fendBuf)
        {
            pkt->endOfPkt = 1;
        }
        else
        {
            pkt->from = pkt->to + strlen((char *)pkt->to) + 1;
            if (pkt->from >= pkt->fendBuf)
            {
                pkt->endOfPkt = 1;
            }
            else
            {
                pkt->subject = pkt->from + strlen((char *)pkt->from) + 1;
                if (pkt->subject >= pkt->fendBuf)
                {
                    pkt->endOfPkt = 1;
                }
                else
                {
                    pkt->text = pkt->subject + strlen((char *)pkt->subject) + 1;
                    if (pkt->text >= pkt->fendBuf)
                    {
                        pkt->endOfPkt = 1;
                    }
                    else
                    {
                        pkt->startValT = pkt->text;
                        pkt->endPtr = pkt->text + strlen((char *)pkt->text);
                        if (pkt->endPtr >= pkt->fendBuf)
                        {
                            pkt->endPtr = NULL;
                            if (pkt->fendBuf < pkt->endBuf)
                            {
                                pkt->endOfPkt = 1;
                            }
                        }
                    }
                }
            }
        }
    }
    if (!pkt->endOfPkt)
    {
        if (memcmp(pkt->startPtr, "\x02\x00", 2) != 0)
        {
            pkt->endOfPkt = 1;
        }
    }
    return (pkt->endPtr == NULL);
}

void pktBegin(PKT *pkt)
{
    pktSetLoc(pkt, pkt->origLoc);
    pkt->startPtr = pkt->startBuf;
    pkt->startValT = pkt->text;
    pkt->endPtr = NULL;
    return;
}
